/* --------------------------------------------------------------------------
 * Copyright (C) 2000-2001 Klar�lvdalens Datakonsult AB. All rights reserved.
 *
 * This file is part of KNUT, the Windows resource converter for Qt GUI
 * Designer. You may only use and modify it according to the license terms
 * appearing in the file LICENSE included in this package.
 *
 * Holders of a commercial license have additional rights as specified in the
 * license terms shipped with the commercial package.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * ------------------------------------------------------------------------ */


/**
 * RES file writer (ListView backend).
 *
 * Source file for class QtListViewWriter.
 */

#include "QtListViewWriter.h"

#include <Resource.h>
#include <ResFile.h>
#include <ResHeader.h>
#include <ResMenu.h>
#include <ResDialog.h>
#include <ResGroupIcon.h>
#include <ResString.h>
#include <ResUserDefined.h>
#include <ResIcon.h>
#include <ResBitmap.h>
#include <ResCursor.h>

#include <qdatetime.h>
#include <qfile.h>
#include <qmap.h>
#include <qnamespace.h>
#include <qobject.h>
#include <qtextstream.h>
#include <qapplication.h>
#include <qlistview.h>
#include <qptrstack.h> //<qstack.h>

QtListViewWriter::QtListViewWriter( QWidget* parent ) :
  previousitem( 0 )
{
  listview = new QListView( parent );
  listview->addColumn( "Name/Number" );
  listview->addColumn( "Type" );
  listview->addColumn( "Info" );
  listview->setSorting( -1 );
  listview->setAllColumnsShowFocus( true );
  listview->setRootIsDecorated( true );
}


QListView* QtListViewWriter::write( const ResourceList* resources )
{
  ResourceListIterator it( *resources );
  Resource* current;
  while( ( current = it.current() ) ) {
	switch( current->header()->numericType() ) {
	case RT_CURSOR:
	  recordCursorResource( current );
	  break;
	case RT_BITMAP:
	  recordBitmapResource( current );
	  break;
	case RT_ICON:
	  recordIconResource( current );
	  break;
	case RT_MENU:
	  recordMenuResource( current );
	  break;
	case RT_DIALOG:
	  recordDialogResource( current );
	  break;
	case RT_STRING:
	  recordStringResource( current );
	  break;
	case RT_FONT:
	  recordFontResource( current );
	  break;
	case RT_FONTDIR:
	  recordFontDirResource( current );
	  break;
	case RT_ACCELERATORS:
	  recordAcceleratorsResource( current );
	  break;
	case RT_RCDATA:
	  recordRCDataResource( current );
	  break;
	case RT_MESSAGETABLE:
	  recordMessageTableResource( current );
	  break;
	case RT_GROUP_CURSOR:
	  recordGroupCursorResource( current );
	  break;
	case RT_GROUP_ICON:
	  recordGroupIconResource( current );
	  break;
	case RT_VERSION:
	  recordVersionResource( current );
	  break;
	case RT_NEWBITMAP:
	  recordNewBitmapResource( current );
	  break;
	case RT_NEWMENU:
	  recordNewMenuResource( current );
	  break;
	case RT_NEWDIALOG:
	  recordNewDialogResource( current );
	  break;
	case RT_TOOLBAR:
	  recordToolbarResource( current );
	  break;
	case RT_DLGINIT:
	  recordDlgInitResource( current );
	  break;
	case RT_ERROR:
	  recordUserDefinedResource( current );
	default:
	  qDebug( "Unknown resource" );
	};
	++it;
  }

  return listview;
}



void QtListViewWriter::recordCursorResource( Resource* current )
{
  previousitem = new QListViewItem( listview, previousitem,
									current->header()->name().toString(),
									"RT_CURSOR" );


  ResCursor* resource = ((ResCursor*)current->data());

  QPixmap iconpix;
  iconpix.convertFromImage( resource->image() );
  previousitem->setPixmap( 2, iconpix );
}


void QtListViewWriter::recordUserDefinedResource( Resource* current )
{
  previousitem = new QListViewItem( listview, previousitem,
									current->header()->name().toString(),
									"User-defined resource" );
}


void QtListViewWriter::recordBitmapResource( Resource* current )
{
  previousitem = new QListViewItem( listview, previousitem,
									current->header()->name().toString(),
									"RT_BITMAP" );

  ResBitmap* resource = ((ResBitmap*)current->data());

  QString infotext;
  infotext.sprintf( "bmp size = %d, dimensions = %d/%d, bit planes = %d, bits per pixel = %d, compression = " + resource->GetCompressionTxt(),
  resource->ms.SizeImage,
  resource->ms.Width,
  resource->ms.Height,
  resource->ms.Planes,
  resource->ms.BitCount );

  previousitem->setText( 2, infotext );

  QPixmap iconpix;
  iconpix.convertFromImage( resource->image() );
  previousitem->setPixmap( 2, iconpix );
}


void QtListViewWriter::recordIconResource( Resource* current )
{
  previousitem = new QListViewItem( listview, previousitem,
									current->header()->name().toString(),
									"RT_ICON" );

  ResIcon* resource = ((ResIcon*)current->data());

  QPixmap iconpix;
  iconpix.convertFromImage( resource->image() );
  previousitem->setPixmap( 2, iconpix );
}


void QtListViewWriter::recordMenuResource( Resource* current )
{
  QStack<QListViewItem> itemstack;

  previousitem = new QListViewItem( listview, previousitem,
									current->header()->name().toString(),
									"RT_MENU" );

  // PENDING(kalle) Remove
  return;

  QListViewItem* previousitemforfollowingresources = previousitem;

  QString infotext;
  ResMenuItemList* menuitems = ((ResMenu*)current->data())->items();
  infotext.sprintf( "%d menu items", menuitems->count() );
  previousitem->setText( 2, infotext );

  QListViewItem* previoussubitem = 0;
  for( uint i = 0; i < menuitems->count(); i++ ) {
	ResMenuItem* item = menuitems->at( i );
	QString infotext;
	infotext.sprintf( "%s (id = %d, grayed = %s, inactive = %s, checked = %s )",
					  item->separator ? "--------" : item->itemtext.latin1(),
					  item->itemid,
					  item->grayed ? "true" : "false",
					  item->inactive ? "true" : "false",
					  item->checked ? "true" : "false" );
	previoussubitem = new QListViewItem( previousitem, previoussubitem, "", "",
										 infotext );
	if( item->popup ) {
	  // this is a popup menu: need to go one level down
	  // first push the old parent on the stack
	  itemstack.push( previousitem );
	  // now make the current item the parent for all following ones until the
	  // old parent is popped from the stack again
	  previousitem = previoussubitem;
	}
	if( item->endmenu ) {
	  // this ends the current menu: need to one level up
	  // pop the old parent from the stack and make the current parent the
	  // previous sub item again
	  previoussubitem = previousitem;
	  previousitem = itemstack.pop();
	}
  }
  // after the loop, the stack should be empty again
  ASSERT( itemstack.isEmpty() );
  previousitem = previousitemforfollowingresources;
}


static QString cdStyleToTypeString( nameorordinal classId )
{
  if( !classId.isordinal )
	return "user-defined";
  else {
	switch( classId.id ) {
	case BUTTON:
	  return "Button";
	case EDIT:
	  return "Edit";
	case STATIC:
	  return "Static";
	case LISTBOX:
	  return "ListBox";
	case SCROLLBAR:
	  return "ScrollBar";
	case COMBOBOX:
	  return "ComboBox";
	}
  }
}


void QtListViewWriter::recordDialogResource( Resource* current )
{
  previousitem = new QListViewItem( listview, previousitem,
									current->header()->name().toString(),
									"RT_DIALOG" );
  ResDialog* data = ((ResDialog*)current->data());
  const DialogBoxHeader* dbheader = data->header();
  QString infotext;
  infotext.sprintf( "x = %d, y = %d, w = %d, h = %d, %d components, caption = %s",
					dbheader->x, dbheader->y, dbheader->cx, dbheader->cy,
					dbheader->numberOfItems, dbheader->caption.latin1() );
  previousitem->setText( 2, infotext );

  // Now the dialog controls
  QListViewItem* previoussubitem = 0;
  for( uint i = 0; i < dbheader->numberOfItems; i++ ) {
	ControlData* item = data->controls()->at( i );
	QString infotext;
	infotext.sprintf( "type = %s, x = %d, y = %d, cx = %d, cy = %d, text = %s",
					  cdStyleToTypeString( item->classId ).data(),
					  item->x, item->y, item->cx, item->cy,
					  item->text.toString().data() );

	previoussubitem = new QListViewItem( previousitem, previoussubitem, "", "",
										 infotext );
  }
}


void QtListViewWriter::recordStringResource( Resource* current )
{
  previousitem = new QListViewItem( listview, previousitem,
									current->header()->name().toString(),
									"RT_STRING" );
  ResString* data = (ResString*)current->data();
  QMap<Q_UINT16,QString>::Iterator it;
  QListViewItem* previoussubitem = 0;
  for( it = data->strings()->begin(); it != data->strings()->end(); ++it )
	previoussubitem = new QListViewItem( previousitem, previoussubitem, "", "",
										 it.data() );

}


void QtListViewWriter::recordFontResource( Resource* current )
{
  previousitem = new QListViewItem( listview, previousitem,
									current->header()->name().toString(),
									"RT_FONT" );
  qDebug( "Sorry, not implemented: QtListViewWriter::recordFontResource" );
}


void QtListViewWriter::recordFontDirResource( Resource* current )
{
  previousitem = new QListViewItem( listview, previousitem,
									current->header()->name().toString(),
									"RT_FONTDIR" );
  qDebug( "Sorry, not implemented: QtListViewWriter::recordFontDirResource" );
}


void QtListViewWriter::recordAcceleratorsResource( Resource* current )
{
  previousitem = new QListViewItem( listview, previousitem,
									current->header()->name().toString(),
									"RT_ACCEL" );
  qDebug( "Sorry, not implemented: QtListViewWriter::recordAcceleratorsResource" );
}


void QtListViewWriter::recordRCDataResource( Resource* current )
{
  previousitem = new QListViewItem( listview, previousitem,
									current->header()->name().toString(),
									"RT_RCDATA" );
  qDebug( "Sorry, not implemented: QtListViewWriter::recordRCDataResource" );
}


void QtListViewWriter::recordMessageTableResource( Resource* current )
{
  previousitem = new QListViewItem( listview, previousitem,
									current->header()->name().toString(),
									"RT_MESSAGETABLE" );
  qDebug( "Sorry, not implemented: QtListViewWriter::recordMessageTableResource" );
}


void QtListViewWriter::recordGroupCursorResource( Resource* current )
{
  previousitem = new QListViewItem( listview, previousitem,
									current->header()->name().toString(),
									"RT_GROUPCURSOR" );
  qDebug( "Sorry, not implemented: QtListViewWriter::recordGroupCursorResource" );
}


void QtListViewWriter::recordGroupIconResource( Resource* current )
{
  previousitem = new QListViewItem( listview, previousitem,
									current->header()->name().toString(),
									"RT_GROUPICON" );
  QString infotext;
  infotext.sprintf( "%d components", ((ResGroupIcon*)current->data())->iconHeader()->wCount );
  previousitem->setText( 2, infotext );

  // Add Info about the icons belonging to this icon group
  QListViewItem* previoussubitem = 0;
  QList<ResourceDirectory>* iconList = ((ResGroupIcon*)current->data())->iconList();
  for( uint i = 0; i < iconList->count(); i++ ) {
	QString infotext;
	infotext.sprintf( "%dx%dx%d, bytes = %d, name = %d",
					  iconList->at( i )->bWidth,
					  iconList->at( i )->bHeight,
					  iconList->at( i )->wBitCount,
					  iconList->at( i )->lBytesInRes,
					  iconList->at( i )->wNameOrdinal );
	previoussubitem = new QListViewItem( previousitem, previoussubitem, "", "",
										 infotext );
  }
}


void QtListViewWriter::recordVersionResource( Resource* current )
{
  previousitem = new QListViewItem( listview, previousitem,
									current->header()->name().toString(),
									"RT_VERSION" );
  qDebug( "Sorry, not implemented: QtListViewWriter::recordVersionResource" );
}


void QtListViewWriter::recordNewBitmapResource( Resource* current )
{
  previousitem = new QListViewItem( listview, previousitem,
									current->header()->name().toString(),
									"RT_NEWBITMAP" );
  qDebug( "Sorry, not implemented: QtListViewWriter::recordNewBitmapResource" );
}


void QtListViewWriter::recordNewMenuResource( Resource* current )
{
  previousitem = new QListViewItem( listview, previousitem,
									current->header()->name().toString(),
									"RT_NEWMENU" );
  qDebug( "Sorry, not implemented: QtListViewWriter::recordNewMenuResource" );
}


void QtListViewWriter::recordNewDialogResource( Resource* current )
{
  previousitem = new QListViewItem( listview, previousitem,
									current->header()->name().toString(),
									"RT_NEWDIALOG" );
  qDebug( "Sorry, not implemented: QtListViewWriter::recordNewDialogResource" );
}


void QtListViewWriter::recordToolbarResource( Resource* current )
{
  previousitem = new QListViewItem( listview, previousitem,
									current->header()->name().toString(),
									"RT_TOOLBAR" );
  qDebug( "Sorry, not implemented: QtListViewWriter::recordToolbarResource" );
}


void QtListViewWriter::recordDlgInitResource( Resource* current )
{
  previousitem = new QListViewItem( listview, previousitem,
									current->header()->name().toString(),
									"RT_DLGINIT" );
  qDebug( "Sorry, not implemented: QtListViewWriter::recordDlgInitResource" );
}



