/* --------------------------------------------------------------------------
 * Copyright (C) 2000-2001 Klarlvdalens Datakonsult AB. All rights reserved.
 *
 * This file is part of KNUT, the Windows resource converter for Qt GUI
 * Designer. You may only use and modify it according to the license terms
 * appearing in the file LICENSE included in this package.
 *
 * Holders of a commercial license have additional rights as specified in the
 * license terms shipped with the commercial package.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * ------------------------------------------------------------------------ */


/**
 * RES file writer (Qt source backend).
 *
 * Source file for class QtSourceWriter.
 */

#include "QtSourceWriter.h"

#include <Resource.h>
#include <ResFile.h>
#include <ResHeader.h>
#include <ResMenu.h>
#include <ResDialog.h>

#include <qdatetime.h>
#include <qfile.h>
#include <qmap.h>
#include <qnamespace.h>
#include <qobject.h>
#include <qtextstream.h>

QtSourceWriter::QtSourceWriter( const QString& filename )
{
  iodevice = new QFile( filename );
  filecreated = true;
}


QtSourceWriter::QtSourceWriter( QIODevice* iodevice )
{
  this->iodevice = iodevice;
  filecreated = false;
}


QtSourceWriter::~QtSourceWriter()
{
  if( filecreated ) {
	iodevice->close();
	delete iodevice;
  }
}


bool QtSourceWriter::write( const ResourceList* resources )
{
  qDebug( "QtSourceWriter::write()" );

  if( !iodevice->isOpen() ) {
	if( !iodevice->open( IO_WriteOnly | IO_Truncate ) )
	  return false;
  }
  if( !iodevice->isWritable() )
	return false;
  
  QTextStream s( iodevice );

  s << "/**************************************************\n";
  s << " * This file was automatically generated by resconv \n";
  s << " * at " << QDateTime::currentDateTime().toString() << "\n";
  s << " * DO NOT EDIT!\n";
  s << " *************************************************/\n\n\n";

  // write headers
  s << "#include <qcursor.h>\n"
	<< "#include <qmenubar.h>\n"
	<< "#include <qpixmap.h>\n"
	<< "#include <qpopupmenu.h>\n"
	<< "#include <qstring.h>\n"
	<< "#include <qdialog.h>\n"
	<< "#include <qfont.h>\n"
	<< "#include <qdir.h>\n"
	<< "#include <qaccel.h>\n"
	<< "#include <qtoolbar.h>\n"
	<< "#include <qstack.h>\n"
	<< "#include <qfont.h>\n"
	<< "#include <qlabel.h>\n"
	<< "#include <qpushbutton.h>\n"
	<< "\n\n";

  s << "class CursorFactory\n"
	<< "{\n"
	<< "public:\n"
	<< "\tstatic QCursor* create( ushort id );\n"
	<< "\tstatic QCursor* create( const QString& );\n"
	<< "};\n\n";

  s << "class PixmapFactory\n"
	<< "{\n"
	<< "public:\n"
	<< "\tstatic QPixmap* create( ushort id );\n"
	<< "\tstatic QPixmap* create( const QString& );\n"
	<< "};\n\n";

  s << "class MenuFactory\n"
	<< "{\n"
	<< "public:\n"
	<< "\tstatic QMenuBar* create( ushort id, QWidget* parent );\n"
	<< "\tstatic QMenuBar* create( const QString&, QWidget* parent );\n"
	<< "};\n\n";

  s << "class DialogFactory\n"
	<< "{\n"
	<< "public:\n"
	<< "\tstatic QDialog* create( ushort id, QWidget* parent, const char* name );\n"
	<< "\tstatic QDialog* create( const QString& id, QWidget* parent, const char* name );\n"
	<< "};\n\n";

  s << "class FontFactory\n"
	<< "{\n"
	<< "public:\n"
	<< "\tstatic QFont* create( ushort id );\n"
	<< "\tstatic QFont* create( const QString& );\n"
	<< "};\n\n";

  s << "class StringFactory\n"
	<< "{\n"
	<< "public:\n"
	<< "\tstatic QString* create( ushort id );\n"
	<< "\tstatic QString* create( const QString& );\n"
	<< "};\n\n";

  s << "class QAccelFactory\n"
	<< "{\n"
	<< "public:\n"
	<< "\tstatic QAccel* create( ushort id );\n"
	<< "\tstatic QAccel* create( const QString& );\n"
	<< "};\n\n";

  s << "class ToolbarFactory\n"
	<< "{\n"
	<< "public:\n"
	<< "\tstatic QToolBar* create( ushort id, QWidget* parent );\n"
	<< "\tstatic QToolBar* create( const QString&, QWidget* parent );\n"
	<< "};\n\n";


  ResourceListIterator it( *resources );
  Resource* current;
  while( ( current = it.current() ) ) {
	switch( current->header()->numericType() ) {
	case RT_CURSOR:
	  recordCursorResource( current );
	  break;
	case RT_BITMAP:
	  recordBitmapResource( current );
	  break;
	case RT_ICON:
	  recordIconResource( current );
	  break;
	case RT_MENU:
	  recordMenuResource( current );
	  break;
	case RT_DIALOG:
	  recordDialogResource( s, current );
	  break;
	case RT_STRING:
	  recordStringResource( current );
	  break;
	case RT_FONT:
	  recordFontResource( current );
	  break;
	case RT_FONTDIR:
	  recordFontDirResource( current );
	  break;
	case RT_ACCELERATORS:
	  recordAcceleratorsResource( current );
	  break;
	case RT_RCDATA:
	  recordRCDataResource( current );
	  break;
	case RT_MESSAGETABLE:
	  recordMessageTableResource( current );
	  break;
	case RT_GROUP_CURSOR:
	  recordGroupCursorResource( current );
	  break;
	case RT_GROUP_ICON:
	  recordGroupIconResource( current );
	  break;
	case RT_VERSION:
	  recordVersionResource( current );
	  break;
	case RT_NEWBITMAP:
	  recordNewBitmapResource( current );
	  break;
	case RT_NEWMENU:
	  recordNewMenuResource( current );
	  break;
	case RT_NEWDIALOG:
	  recordNewDialogResource( current );
	  break;
	case RT_TOOLBAR:
	  recordToolbarResource( current );
	  break;
	case RT_DLGINIT:
	  recordDlgInitResource( current );
	  break;
	default:
	  qDebug( "Unknown resource" );
	};
	++it;
  }

  
  // the factory implementations
  s << "struct savedmenuentry {\n"
	<< "\tQMenuData* menu;\n"
	<< "\tQString text;\n"
	<< "\tint id;\n"
	<< "};\n\n\n";
  s << "QMenuBar* MenuFactory::create( ushort id, QWidget* parent )\n"
	<< "{\n"
	<< "\tswitch( id ) {\n";
  ResourceMapByID::Iterator menuit;
  for( menuit = menulist1.begin(); menuit != menulist1.end(); ++menuit ) {
	s << "\tcase " << menuit.key() << ": {\n";
	s << menuit.data();
	s << "\t\tbreak;\n"
	  << "\t}";
  }
  s << "\tdefault:\n"
	<< "\t\tqDebug( \"Unknown menu id\" );\n"
	<< "\t}\n"
	<< "}\n";

  s << "QMenuBar* MenuFactory::create( const QString& id, QWidget* parent )\n"
	<< "{\n";
  ResourceMapByString::Iterator menuit2;
  bool first = true;
  for( menuit2 = menulist2.begin(); menuit2 != menulist2.end(); ++menuit2 ) {
	if( first ) {
	  s << "\tif( id == " << menuit2.key() << " ) {\n";
	  first = false;
	} else
	  s << "\t} else if( id == " << menuit2.key() << " ) {\n";
	s << menuit2.data();
  }
  if( !first )
	s << "\t} else\n";
  s << "\t\tqDebug( \"Unknown menu id\" );\n"
	<< "}\n\n";

  s << "QDialog* DialogFactory::create( ushort id, QWidget* parent, const char* name )\n";
  s << "{\n"
	<< "\tswitch( id ) {\n";
  ResourceMapByID::Iterator dialogit;
  for( dialogit = dialoglist1.begin(); dialogit != dialoglist1.end();
	   ++dialogit ) {
	s << "\tcase " << dialogit.key() << ": {\n";
	s << dialogit.data();
	s << "\t\tbreak;\n"
	  << "\t}";
  }
  s << "\tdefault:\n"
	<< "\t\tqDebug( \"Unknown dialog id\" );\n"
	<< "\t}\n"
	<< "}\n\n\n";

  s << "QDialog* DialogFactory::create( const QString& id, QWidget* parent,"
       " const char* name )\n"
	<< "{\n";
  ResourceMapByString::Iterator dialogit2;
  first = true;
  for( dialogit2 = dialoglist2.begin(); dialogit2 != dialoglist2.end(); ++dialogit2 ) {
	if( first ) {
	  s << "\tif( id == " << dialogit2.key() << " ) {\n";
	  first = false;
	} else
	  s << "\t} else if( id == " << dialogit2.key() << " ) {\n";
	s << dialogit2.data();
  }
  if( !first )
	s << "\t} else\n";
  s << "\t\tqDebug( \"Unknown dialog id\" );\n"
	<< "}\n\n";
	
  return true;
}


void QtSourceWriter::fixItemText( QString& text, int& accel )
{
  QString origtext = text;

  // find a tab
  int tabpos = text.findRev( '\t' );
  if( tabpos == -1 ) // no tab - we are done
	return;

  // save the substring right of the tab
  QString accelstr = text.mid( tabpos+1 );
  // cut the string to everything left of the tab
  text = text.left( tabpos );
  
  // look for a plus sing
  int pluspos = accelstr.find( '+' );
  if( pluspos != -1 ) {
	// there is a plus sign -> the string is of the STRG+xxx or ALT+xxx
	
	// first cover the part to the left of the plus (look for English, German and localized)
	if( accelstr.left( pluspos ).upper() == "CTRL" || 
		accelstr.left( pluspos ).upper() == "STRG" ||
		accelstr.left( pluspos ).upper() == QObject::tr( "CTRL" ) )
	  accel += Qt::CTRL;
	else if( accelstr.left( pluspos ).upper() == "ALT" ||
			 accelstr.left( pluspos ).upper() == QObject::tr( "ALT" ) )
	  accel += Qt::ALT;

	// Now the part to the right. If it is one letter long, we take it
	// directly, if it is two letters long and the first one is (f|F), or it
	// is three letters long and the first one is (f|F),
	// we interpret it as a function key. If it is ENTER/EINGABETASTE,
	// we use that. Otherwise, we give up.
	QString rightplus = accelstr.mid( pluspos + 1 );
	if( rightplus.length() == 1 ) {
	  QChar c = accelstr.at( accelstr.length() - 1 );
	  accel += c.latin1();
	} else if( rightplus.length() == 2 &&
			   ( rightplus.at( 0 ) == 'F' ) ||
			   ( rightplus.at( 0 ) == 'f' ) ) {
	  QChar c = accelstr.at( accelstr.length() - 1 );
	  int digitvalue = c.digitValue();
	  if( digitvalue == -1 ) {
		qDebug( "Sorry, cannot parse accelerator in %s", origtext.latin1() );
		accel = 0;
	  } else 
		accel += Qt::Key_F1 + digitvalue - 1;
	} else if( rightplus.length() == 3 &&
			   ( rightplus.at( 0 ) == 'F' ) ||
			   ( rightplus.at( 0 ) == 'f' ) ) {
	  QString numstr = accelstr.right( 2 );
	  bool ok;
	  int val = numstr.toInt( &ok );
	  if( ok && val <= 35 ) {
		accel += Qt::Key_F1 + val - 1;
	  } else {
		qDebug( "Sorry, cannot parse accelerator in %s", origtext.latin1() );
		accel = 0;
	  }
	} else if( rightplus.upper() == "ENTER" ||
			   rightplus.upper() == "EINGABETASTE" ||
			   rightplus.upper() == QObject::tr( "ENTER" ) ) {
	  accel += Qt::Key_Enter;
	} else {
	  qDebug( "Sorry, cannot parse accelerator in %s", origtext.latin1() );
	  accel = 0;
	}
  } else {
	// There is no plus sign. All that we understand in this case are function
	// keys.
	if( text.length() == 2 &&
		( text.at( 0 ) == 'F' ) ||
		( text.at( 0 ) == 'f' ) ) {
	  QChar c = text.at( 1 );
	  int digitvalue = c.digitValue();
	  if( digitvalue == -1 ) {
		qDebug( "Sorry, cannot parse accelerator in %s", origtext.latin1() );
		accel = 0;
	  } else 
		accel += Qt::Key_F1 + digitvalue - 1;
	} else if( text.length() == 3 &&
			   ( text.at( 0 ) == 'F' ) ||
			   ( text.at( 0 ) == 'f' ) ) {
	  QString numstr = text.right( 2 );
	  bool ok;
	  int val = numstr.toInt( &ok );
	  if( ok && val <= 35 ) {
		accel += Qt::Key_F1 + val - 1;
	  } else {
		qDebug( "Sorry, cannot parse accelerator in %s", origtext.latin1() );
		accel = 0;
	  }
	}
  }
}


void QtSourceWriter::recordCursorResource( Resource* current )
{
  qDebug( "Sorry, not implemented: QtSourceWriter::recordCursorResource" );
}


void QtSourceWriter::recordBitmapResource( Resource* current )
{
  qDebug( "Sorry, not implemented: QtSourceWriter::recordBitmapResource" );
}


void QtSourceWriter::recordIconResource( Resource* current )
{
  qDebug( "Sorry, not implemented: QtSourceWriter::recordIconResource" );
}


void QtSourceWriter::recordMenuResource( Resource* current )
{
  QString text;
  text += "\t\tQMenuBar* menu = new QMenuBar( parent, ";
  text += '\"' + current->header()->name().toString() + '\"';
  text += " );\n";
  text += "\t\tQStack<savedmenuentry> menustack;\n";
  text += "\t\tQMenuData* currentmenu = menu;\n";

  ResMenu* data = (ResMenu*)current->data();
  ResMenuItemListIterator it( *data->items() );
  ResMenuItem* currentitem;
  int count = 0;
  while( ( currentitem = it.current() ) ) {
	count++;
	// PENDING(kalle) Handle accelerators
	// PENDING(kalle) Handle pixmaps in menus
	// PENDING(kalle) Handle other flags where possible
	// PENDING(kalle) Handle popup menus

	// if this is a popup menu, just push the data on the stack 
	// and start a new one
	if( currentitem->popup ) {
	  text += "\t\tsavedmenuentry* s";
	  text += QString::number( count );
	  text += " = new savedmenuentry;\n";
	  text += "\t\ts";
	  text += QString::number( count );
	  text += "->menu = currentmenu;\n";
	  text += "\t\ts";
	  text += QString::number( count );
	  text += "->text = \"";
	  text += currentitem->itemtext;
	  text += "\";\n";
	  text += "\t\ts";
	  text += QString::number( count );
	  text += "->id = ";
	  text += QString::number( (int)currentitem->itemid );
	  text += ";\n";
	  text += "\t\tmenustack.push( s";
	  text += QString::number( count );
	  text += " );\n";
	  text += QString( "\t\tQPopupMenu* menu" );
	  text += QString::number( count );
	  text += QString( " = new QPopupMenu;\n" );
	  text += QString( "\t\tcurrentmenu = menu" );
	  text += QString::number( count ) += QString( ";\n" );
	} else {
	  if( currentitem->separator )
		text += "\t\tcurrentmenu->insertSeparator();\n";
	  else {
		int accel = 0;
		fixItemText( currentitem->itemtext, accel );
		text += "\t\tcurrentmenu->insertItem( \"" + currentitem->itemtext + "\", " +
		  QString::number( (int)currentitem->itemid ) + " );\n";
		if( accel )
		  text += "\t\tcurrentmenu->setAccel( " +  QString::number( accel ) + ", " + QString::number(
																  (int)currentitem->itemid ) + " );\n";
		if( currentitem->grayed || currentitem->inactive )
		  text += QString( "\t\tcurrentmenu->setItemEnabled( " ) + QString::number( (int)currentitem->itemid ) + ", " +
			"false );\n";
		if( currentitem->checked )
		  text += QString( "\t\tcurrentmenu->setItemChecked( " ) + QString::number( (int)currentitem->itemid ) + ", " +
			"true );\n";
	  }
	  
	  if( currentitem->endmenu ) {
		text += "\t\tsavedmenuentry* s";
		text += QString::number( count );
		text += " = menustack.pop();\n";
		text += "\t\ts";
		text += QString::number( count );
		text += "->menu->insertItem( s";
		text += QString::number( count );
		text += "->text, (QPopupMenu*)currentmenu, s";
		text += QString::number( count );
		text += "->id );\n";
		text += "\t\tcurrentmenu = s";
		text += QString::number( count );
		text += "->menu;\n";
		text += "\t\tdelete s";
		text += QString::number( count );
		text += ";\n";
	  }
	}
	++it;
  }

  text += "\t\treturn menu;\n";
  
  if( current->header()->name().isordinal )
	menulist1.insert( current->header()->name().id, text );
  else
	menulist2.insert( current->header()->name().string, text );
}


void QtSourceWriter::recordDialogResource( QTextStream& str, 
										   Resource* current )
{
  ResDialog* data = (ResDialog*)current->data();

  QString dialogname = "Dialog_" + current->header()->name().toString();

  // dialog class declaration, directly written to file
  str << "class " << dialogname << " : public QDialog\n";
  str << "{\n";
  str << "\tQ_OBJECT\n";
  str << "public:\n";
  str << "\t" << dialogname << "( QWidget* parent = 0, const char* name = 0 );\n\n";
  str << "private:\n";
  str << "};\n\n";
  
  // constructor implementation, directly written to file
  str << dialogname << "::" << dialogname << "( QWidget* parent, const char* name ):\n";
  str << "\tQDialog( parent, name, true )\n";
  str << "{\n";
  // the caption
  str << "\tsetCaption( tr( \"" << data->header()->caption << "\" ) );\n\n";

  // the font
  str << "\tsetFont( QFont( \"" << data->header()->fontName << "\", " << 
	data->header()->wPointSize << " ) );\n";

  // size and position
  str << "\tsetGeometry( " << data->header()->x << ", " 
	  << data->header()->y << ", " << data->header()->cx << ", "
	  << data->header()->cy << " );\n\n";

  // PENDING(kalle) Handle styles and extended styles.
  
  // Now do the controls
  QListIterator<ControlData> it( *data->controls() );
  ControlData* currentcd;
  while( ( currentcd = it.current() ) ) {
	++it;
	writeControl( str, currentcd );
  }


  str << "}\n\n\n";

  
  // factory implementation, saved for later
  QString text;
  QTextStream ts( &text, IO_WriteOnly );
  ts << "\t\treturn new " << dialogname << "( parent, name );";

  if( current->header()->name().isordinal )
	dialoglist1.insert( current->header()->name().id, text );
  else
	dialoglist2.insert( current->header()->name().string, text );
}


void QtSourceWriter::recordStringResource( Resource* current )
{
  qDebug( "Sorry, not implemented: QtSourceWriter::recordStringResource" );
}


void QtSourceWriter::recordFontResource( Resource* current )
{
  qDebug( "Sorry, not implemented: QtSourceWriter::recordFontResource" );
}


void QtSourceWriter::recordFontDirResource( Resource* current )
{
  qDebug( "Sorry, not implemented: QtSourceWriter::recordFontDirResource" );
}


void QtSourceWriter::recordAcceleratorsResource( Resource* current )
{
  qDebug( "Sorry, not implemented: QtSourceWriter::recordAcceleratorsResource" );
}


void QtSourceWriter::recordRCDataResource( Resource* current )
{
  qDebug( "Sorry, not implemented: QtSourceWriter::recordRCDataResource" );
}


void QtSourceWriter::recordMessageTableResource( Resource* current )
{
  qDebug( "Sorry, not implemented: QtSourceWriter::recordMessageTableResource" );
}


void QtSourceWriter::recordGroupCursorResource( Resource* current )
{
  qDebug( "Sorry, not implemented: QtSourceWriter::recordGroupCursorResource" );
}


void QtSourceWriter::recordGroupIconResource( Resource* current )
{
  qDebug( "Sorry, not implemented: QtSourceWriter::recordGroupIconResource" );
}


void QtSourceWriter::recordVersionResource( Resource* current )
{
  qDebug( "Sorry, not implemented: QtSourceWriter::recordVersionResource" );
}


void QtSourceWriter::recordNewBitmapResource( Resource* current )
{
  qDebug( "Sorry, not implemented: QtSourceWriter::recordNewBitmapResource" );
}


void QtSourceWriter::recordNewMenuResource( Resource* current )
{
  qDebug( "Sorry, not implemented: QtSourceWriter::recordNewMenuResource" );
}


void QtSourceWriter::recordNewDialogResource( Resource* current )
{
  qDebug( "Sorry, not implemented: QtSourceWriter::recordNewDialogResource" );
}


void QtSourceWriter::recordToolbarResource( Resource* current )
{
  qDebug( "Sorry, not implemented: QtSourceWriter::recordToolbarResource" );
}


void QtSourceWriter::recordDlgInitResource( Resource* current )
{
  qDebug( "Sorry, not implemented: QtSourceWriter::recordDlgInitResource" );
}

  
void QtSourceWriter::writeControl( QTextStream& s, ControlData* cd )
{
  static int ctrlno = 0;
  ctrlno++;
  QString nostr = QString::number( ctrlno );
  // PENDING(kalle) Support more styles
  if( !cd->classId.isordinal ) {
	qDebug( "Sorry, User-defined controls are not and will never be supported" );
	return;
  }

  switch( cd->classId.id ) {
  case BUTTON:
	if( cd->lStyle & BS_PUSHBUTTON ||
		cd->lStyle & BS_DEFPUSHBUTTON ) {
	  // push button
	  s << "\tQPushButton* pb" << nostr << "PB = new QPushButton( tr( \""
		<< cd->text.toString() << "\" ), this );\n";
	  s << "\tpb" << nostr << "PB->setGeometry( " << cd->x << ", " << cd->y
		<< ", " << cd->cx << ", " << cd->cy << " );\n";
	  if( cd->lStyle & BS_DEFPUSHBUTTON )
		s << "\tpb" << nostr << "PB->setDefault( true );\n";
	} else if( cd->lStyle & BS_CHECKBOX ) {
	  s << "\tQCheckBox* cb" << nostr << "CB = new QCheckBox( tr( \""
		<< cd->text.toString() << "\" ), this );\n";
	  s << "\tcb" << nostr << "CB->setGeometry( " << cd->x << ", " << cd->y
		<< ", " << cd->cx << ", " << cd->cy << " );\n";
	} else if( cd->lStyle & BS_GROUPBOX ) {
 	  s << "\tQGroupBox* gb" << nostr << "GB = new QGroupBox( tr( \""
		<< cd->text.toString() << "\" ), this );\n";
	  s << "\tgb" << nostr << "GB->setGeometry( " << cd->x << ", " << cd->y
		<< ", " << cd->cx << ", " << cd->cy << " );\n";
	} else if( cd->lStyle & BS_RADIOBUTTON ) {
 	  s << "\tQRadioButton* rb" << nostr << "RB = new QRadioButton( tr( \""
		<< cd->text.toString() << "\" ), this );\n";
	  s << "\trb" << nostr << "RB->setGeometry( " << cd->x << ", " << cd->y
		<< ", " << cd->cx << ", " << cd->cy << " );\n";
	} else if( cd->lStyle & BS_AUTOCHECKBOX ||
			   cd->lStyle & BS_AUTO3STATE ||
			   cd->lStyle & BS_AUTORADIOBUTTON ||
			   //			   cd->lStyle & BS_PUSHBOX ||
			   cd->lStyle & BS_3STATE )
	  qDebug( "Sorry, unimplemented Button-like control" );
  case EDIT:  
	qDebug( "Sorry, not implemented: Edit-like controls" );
	break;
  case STATIC: {
	if( cd->lStyle & SS_ICON ) // ICON resource
	  qDebug( "Sorry, not implemented: Icons" );
	else {
	  // labels
	  s << "\tQLabel* l" << nostr << "LA = new QLabel( tr( \""
		<< cd->text.toString() << "\" ), this );\n";
	  s << "\tl" << nostr << "LA->setGeometry( " << cd->x << ", " << cd->y 
		<< ", " << cd->cx << ", " << cd->cy << " );\n";
	  if( cd->lStyle & ES_RIGHT ) // RTEXT resource
		s << "\tl" << nostr << "LA->setAlignment( AlignRight );\n\n";
	  else if( cd->lStyle & ES_CENTER ) // CTEXT resource
		s << "\tl" << nostr << "LA->setAlignment( AlignCenter );\n\n";
	  else // LTEXT resource
		s << "\tl" << nostr << "LA->setAlignment( AlignLeft );\n\n";
	}
	break;
  }
  case LISTBOX:
	qDebug( "Sorry, not implemented: Listbox-like controls" );
	break;
  case SCROLLBAR:
	qDebug( "Sorry, not implemented: Scrollbar-like controls" );
	break;
  case COMBOBOX:
	qDebug( "Sorry, not implemented: Combobox-like controls" );
	break;
  }
}

