/* --------------------------------------------------------------------------
 * Copyright (C) 2000-2001 Klar�lvdalens Datakonsult AB. All rights reserved.
 *
 * This file is part of KNUT, the Windows resource converter for Qt GUI
 * Designer. You may only use and modify it according to the license terms
 * appearing in the file LICENSE included in this package.
 *
 * Holders of a commercial license have additional rights as specified in the
 * license terms shipped with the commercial package.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * ------------------------------------------------------------------------ */


/**
 * RES file writer (Qt Designer XML backend).
 *
 * Source file for class QtDesignerWriter.
 */

#include "QtDesignerWriter.h"

#include <Resource.h>
#include <ResFile.h>
#include <ResHeader.h>
#include <ResMenu.h>
#include <ResDialog.h>
#include <ResGroupIcon.h>
#include <ResIcon.h>

#include <qdatetime.h>
#include <qfile.h>
#include <qdir.h>
#include <qmap.h>
#include <qnamespace.h>
#include <qobject.h>
#include <qtextstream.h>
#include <qbuffer.h>
#include <qptrlist.h>//<qlist.h>
#include <qfontmetrics.h>
#include <qregexp.h>


#include <qpainter.h>
#include <qpixmap.h>
#include <qbitmap.h>
#include <qimage.h>


#include <zlib.h>

//#include "iostream.h"


QtDesignerWriter::QtDesignerWriter ( const QString& filename, float scalefactor,
                                     const QString& rcfilename, QProgressDialog *pdlg ) {
  iodevice = new QFile ( filename );
  filecreated = true;
  this->scalefactor = scalefactor;
  InitFontAndCtrlMatcher ( rcfilename );
  dlg = pdlg;
}


QtDesignerWriter::QtDesignerWriter ( QIODevice* iodevice, float scalefactor,
                                     const QString& rcfilename, QProgressDialog *pdlg ) {
  this->iodevice = iodevice;
  filecreated = false;
  this->scalefactor = scalefactor;
  InitFontAndCtrlMatcher ( rcfilename );
  dlg = pdlg;
}


QtDesignerWriter::~QtDesignerWriter() {
  if ( filecreated ) {
    iodevice->close();
    delete iodevice;
  }
}


class FntMatchEntry {
  public:
    FntMatchEntry ( const QString& rName, const QString& rNewName, const float& rMult ) {
      sOrg = rName;
      sNew = rNewName;
      nMult = rMult;
    }
    QString GetName()    const  {
      return sOrg;
    }
    QString GetNewName() const  {
      return sNew;
    }
    float   GetMult()    const  {
      return nMult;
    }
  private:
    QString sOrg;
    QString sNew;
    float   nMult;
};


class CtrlMatchEntry {
  public:
    CtrlMatchEntry ( const QString& rName, const WORD nId, const Q_UINT32 nStyle ) {
      sOrg = rName;
      nNewId = nId;
      nNewStyle = nStyle;
    }
    QString  GetName()     const  {
      return sOrg;
    }
    WORD     GetNewId()    const  {
      return nNewId;
    }
    Q_UINT32 GetNewStyle() const  {
      return nNewStyle;
    }
  private:
    QString  sOrg;
    WORD     nNewId;
    Q_UINT32 nNewStyle;
};


void QtDesignerWriter::InitFontAndCtrlMatcher ( const QString& rcfilename ) {
  FontMatcher.setAutoDelete ( TRUE );
  CtrlMatcher.setAutoDelete ( TRUE );
  int nFont = 0;
  int nCtrl = 0;
  if ( rcfilename.length() ) {
    QFile f ( rcfilename );
    if ( f.open ( IO_ReadOnly ) ) {
      QTextStream t ( &f );
      QString  sOrg;
      QString  sNew;
      float    nMulti;
      WORD     nId;
      Q_UINT32 nStyle;
      QString  s;
      QString  s2;
      bool     bFontSection = false;
      bool     bCtrlSection = false;
      bool     bOk;
      while ( !t.eof() ) {
        s = t.readLine();
        if ( 0 == s.stripWhiteSpace().find ( '[' ) ) {
          s2 = s.simplifyWhiteSpace().upper();
          if ( 0 == s2.compare ( "[REPLACE FONTS]" ) ) {
            bFontSection = true;
            bCtrlSection = false;
          } else if ( 0 == s2.compare ( "[REPLACE CONTROLS]" ) ) {
            bFontSection = false;
            bCtrlSection = true;
          } else {
            bFontSection = false;
            bCtrlSection = false;
          }
        } else if ( bFontSection || bCtrlSection ) {
          int nP1 =           s.find ( ':', 1 );
          int nP2 = 0 < nP1 ? s.find ( ':', 1 + nP1 ) : 0;
          if ( ( bFontSection  &&  0 < nP1
                 &&      nP1 < nP2 - 1
                 &&            nP2 < static_cast<int> ( s.length() ) - 1 )
               || ( bCtrlSection  &&  0 < nP1
                    &&      nP1 < static_cast<int> ( s.length() ) - 1 ) ) {
            sOrg = s.left ( nP1 ).upper();
            if ( bFontSection ) {
              sNew = s.mid ( nP1 + 1, nP2 - nP1 - 1 );
              s2 = s.right ( s.length() - nP2 - 1 );
              nMulti = s2.toFloat ( &bOk );
            } else {
              bOk = true;
              sNew = s.right ( s.length() - nP1 - 1 ).upper();
              // don't use simplifyWhiteSpace() but remove ALL white space:
              s.replace ( QRegExp ( "\\s" ), "" );

              if ( sNew == "BUTTON" ) {
                nId    = BUTTON;
                nStyle = BS_PUSHBUTTON;
              } else if ( sNew == "CHECKBOX" ) {
                nId    = BUTTON;
                nStyle = BS_CHECKBOX;
              } else if ( sNew == "AUTOCHECKBOX" ) {
                nId    = BUTTON;
                nStyle = BS_AUTOCHECKBOX;
              } else if ( sNew == "3STATEBOX" ) {
                nId    = BUTTON;
                nStyle = BS_3STATE;
              } else if ( sNew == "AUTO3STATEBOX" ) {
                nId    = BUTTON;
                nStyle = BS_AUTO3STATE;
              } else if ( sNew == "GROUPBOX" ) {
                nId    = BUTTON;
                nStyle = BS_GROUPBOX;
              } else if ( sNew == "RADIOBUTTON" ) {
                nId    = BUTTON;
                nStyle = BS_RADIOBUTTON;
              } else if ( sNew == "AUTORADIOBUTTON" ) {
                nId    = BUTTON;
                nStyle = BS_AUTORADIOBUTTON;
              } else if ( ( sNew == "EDIT" ) || ( sNew == "EDITLEFT" ) ) {
                nId    = EDIT;
                nStyle = ES_LEFT;
              } else if ( sNew == "EDITRIGHT" ) {
                nId    = EDIT;
                nStyle = ES_RIGHT;
              } else if ( sNew == "EDITCENTER" ) {
                nId    = EDIT;
                nStyle = ES_CENTER;
              } else if ( sNew == "STATIC" ) {
                nId    = STATIC;
                nStyle = 0;
              } else if ( sNew == "LISTBOX" ) {
                nId    = LISTBOX;
                nStyle = 0;
              } else if ( sNew == "SCROLLBAR" ) {
                nId    = SCROLLBAR;
                nStyle = 0;
              } else if ( sNew == "COMBOBOX" ) {
                nId    = COMBOBOX;
                nStyle = 0;
              } else
                bOk = false;
            }
            if ( bOk ) {
              if ( bFontSection ) {
                FontMatcher.append ( new FntMatchEntry ( sOrg, sNew, nMulti ) );
                ++nFont;
              }
              if ( bCtrlSection ) {
                CtrlMatcher.append ( new CtrlMatchEntry ( sOrg, nId, nStyle ) );
                ++nCtrl;
              }
            }
          }
        }
      }
      f.close();
    }
  }
  if ( !nFont )
    FontMatcher.append ( new FntMatchEntry ( "MS Shell Dlg", "helvetica", 0.8 ) );
  if ( !nCtrl ) {
    CtrlMatcher.append ( new CtrlMatchEntry ( "BORSHADE", BUTTON, BS_GROUPBOX ) );
    CtrlMatcher.append ( new CtrlMatchEntry ( "SYSLISTVIEW32", LISTBOX, 0 ) );
  }
}


int QtDesignerWriter::scale ( int value ) {
  float fvalue = ( float ) value;
  fvalue *= scalefactor;
  return ( int ) fvalue;
}


int QtDesignerWriter::unscale ( int value ) {
  float fvalue = ( float ) value;
  fvalue /= scalefactor;
  return ( int ) fvalue;
}


bool QtDesignerWriter::GetFontMatch ( QString font, QString& rOurName,
                                      float&   rOurMult ) {
  QString upFont = font.upper();
  FntMatchEntry* pMatch;
  for ( pMatch = FontMatcher.first(); 0 != pMatch; pMatch = FontMatcher.next() ) {
    if ( 0 == pMatch->GetName().compare ( upFont ) ) {
      rOurName = pMatch->GetNewName();
      rOurMult = pMatch->GetMult();
      return true;
    }
  }
  return false;
}


bool QtDesignerWriter::GetCtrlMatch ( QString style, WORD&     rOurId,
                                      Q_UINT32& rOurStyle ) {
  QString upStyle = style.upper();
  CtrlMatchEntry* pMatch;
  for ( pMatch = CtrlMatcher.first(); 0 != pMatch; pMatch = CtrlMatcher.next() ) {
    if ( 0 == pMatch->GetName().compare ( upStyle ) ) {
      rOurId = pMatch->GetNewId();
      // set Style - but prevent from erasing 'def button' bit!
      if ( ! ( ( BUTTON           == rOurId )
               && ( BS_PUSHBUTTON    == pMatch->GetNewStyle() )
               && ( BS_DEFPUSHBUTTON == ( rOurStyle & 0x0000000FL ) ) ) )
        rOurStyle = ( rOurStyle & 0xFFFFFFF0 ) | pMatch->GetNewStyle();
      return true;
    }
  }
  return false;
}


/*!
  Helper Class for waiting in programms without a.exec() loop.
*/
class MyWaiter : public QObject {
  public:
    MyWaiter ( QObject *parent = 0, const char *name = 0 );
    void wait ( const int msec = 1000 );
  protected:
    void  timerEvent ( QTimerEvent * );
  private:
    int  idT;
    bool bElapsed;
};

MyWaiter::MyWaiter ( QObject *parent, const char *name )
    : QObject ( parent, name ) {
  idT = 0;
  bElapsed = false;
}

void MyWaiter::wait ( const int msec ) {
  idT = startTimer ( msec );
  do
    qApp->processEvents();
  while ( !bElapsed );
  killTimer ( idT );
}

void MyWaiter::timerEvent ( QTimerEvent *e ) {
  if ( idT == e->timerId() )
    bElapsed = true;
}
/*!
  Waits for msec milliseconds even if there is NO
  a.exec() loop running
*/
static void MyWait ( const int msec ) {
  MyWaiter* waiter = new MyWaiter();
  waiter->wait ( msec );
  delete waiter;
}


bool makeTextFileCopy ( const QString& src, const QString& dest ) {
  QFile input ( src );
  if ( !input.open ( IO_ReadOnly ) ) {
    qDebug ( QString ( "ERROR in makeCopy(%1, %2): cannot open %1" ).arg ( src ).arg ( dest ) );
    return false;
  }
  QFile output ( dest );
  if ( !output.open ( IO_WriteOnly ) ) {
    qDebug ( QString ( "ERROR in makeCopy(%1, %2): cannot create %2" ).arg ( src ).arg ( dest ) );
    return false;
  }
  QTextStream inputStream ( &input );
  QTextStream outputStream ( &output );
  outputStream << inputStream.read();
  output.close();
  input.close();
  return true;
}


bool QtDesignerWriter::write ( const ResourceList* resources ) {
  int dialogno = 0;
  if ( !iodevice->isOpen() ) {
    if ( !iodevice->open ( IO_WriteOnly | IO_Truncate ) )
      return false;
  }
  if ( !iodevice->isWritable() )
    return false;

  // make output directory
  QString homeName ( QDir::home().absPath() );
  QString outDirName ( homeName + "/knutdata" );
  QDir outDir ( outDirName );
  if ( !outDir.exists() && !outDir.mkdir ( outDirName ) )
    outDirName = homeName;

  QTextStream ts ( iodevice );

  ResourceListIterator it ( *resources );
  Q_UINT16 nDlgStep = dlg
                      ? it.count() / ( dlg->totalSteps() - dlg->progress() )
                      : 0;
  Q_UINT16 iDlgStep = 1;
  Q_UINT16 iEntry   = 0;
  Resource* current;
  while ( ( current = it.current() ) ) {
//     qDebug("while");
    if ( dlg ) {
      iEntry++;
      QString s ( "converting resource #" );
      s += QString::number ( iEntry );
      dlg->setLabelText ( s );
      if ( iDlgStep > nDlgStep ) {
        dlg->setProgress ( dlg->progress() + 1 );
        iDlgStep = 1;
      } else
        iDlgStep++;
      qApp->processEvents();
    }
//     qDebug("switch");
    switch ( current->header()->numericType() ) {
      case RT_CURSOR: {
//         qDebug("RT_CURSOR");
        recordCursorResource ( current );
        break;
      }
      case RT_BITMAP: {
//         qDebug("RT_BITMAP");
        recordBitmapResource ( current );
        break;
      }
      case RT_ICON: {
//         qDebug("RT_ICON");
        recordIconResource ( current );
        break;
      }
      case RT_MENU: {
//         qDebug("RT_MENU");
        recordMenuResource ( current );
        break;
      }
      case RT_DIALOG: {
        qDebug("RT_DIALOG");
        imagesfordialog.clear();

        // generate a filename
        dialogno++;
        QString dlgfilename = QString ( "%1/dialog_%2.ui" ).arg ( outDirName ).arg ( dialogno );
        qDebug(dlgfilename);
        // Create the dialog file and open a stream for it.
        QFile dlgfile ( dlgfilename );
        if ( dlgfile.open ( IO_WriteOnly | IO_Truncate ) ) {
          QTextStream dlgstream ( &dlgfile );

          // write the dialog filename to the central file
          ts << dlgfilename << endl;

          QString classDeclaration = QString ( "<class>Form%1</class>" ).arg ( dialogno );
//        dlgstream << "<!DOCTYPE UI><UI version=\"3.0\" stdsetdef=\"1\">" << endl;
          dlgstream << "<!DOCTYPE UI><UI>" << endl;
          dlgstream << classDeclaration << endl;
          recordDialogResource ( dlgstream, current );
          recordImagesForDialog ( dlgstream, current );
          dlgstream << "</UI>" << endl;

          // close the dialog file
          dlgstream.unsetDevice();
          dlgfile.close();
          makeTextFileCopy (
            dlgfilename,
            QString ( "%1/dialog_%2_knut.ui" ).arg ( outDirName ).arg ( dialogno ) );
        }
        break;
      }
      case RT_STRING: {
//         qDebug("RT_STRING");
        recordStringResource ( current );
        break;
      }
      case RT_FONT: {
//         qDebug("RT_FONT");
        recordFontResource ( current );
        break;
      }
      case RT_FONTDIR: {
//         qDebug("RT_FONTDIR");
        recordFontDirResource ( current );
        break;
      }
      case RT_ACCELERATORS: {
//         qDebug("RT_ACCELERATORS");
        recordAcceleratorsResource ( current );
        break;
      }
      case RT_RCDATA: {
//         qDebug("RT_RCDATA");
        recordRCDataResource ( current );
        break;
      }
      case RT_MESSAGETABLE: {
//         qDebug("RT_MESSAGETABLE");
        recordMessageTableResource ( current );
        break;
      }
      case RT_GROUP_CURSOR: {
//         qDebug("RT_GROUP_CURSOR");
        recordGroupCursorResource ( current );
        break;
      }
      case RT_GROUP_ICON: {
//         qDebug("RT_GROUP_ICON");
        recordGroupIconResource ( current );
        break;
      }
      case RT_VERSION: {
//         qDebug("RT_VERSION");
        recordVersionResource ( current );
        break;
      }
      case RT_NEWBITMAP: {
//         qDebug("RT_NEWBITMAP");
        recordNewBitmapResource ( current );
        break;
      }
      case RT_NEWMENU: {
//         qDebug("RT_NEWMENU");
        recordNewMenuResource ( current );
        break;
      }
      case RT_NEWDIALOG: {
//         qDebug("RT_NEWDIALOG");
        recordNewDialogResource ( current );
        break;
      }
      case RT_TOOLBAR: {
//         qDebug( "RT_TOOLBAR");
        recordToolbarResource ( current );
        break;
      }
      case RT_DLGINIT: {
//         qDebug( "RT_DLGINIT");
        recordDlgInitResource ( current );
        break;
      }
      default:{
        qDebug ( "Unknown resource" );
        break;
      }
    };
    ++it;
    if ( it.current() == NULL || *it == 0)
      break;
//     qDebug("next step");
  }
  dlg->close();
// qDebug("caboom");
  if ( dlg && !dialogno ) {
    qDebug("oops");
    dlg->setLabelText ( "Sorry, no dialog resources found." );
    dlg->setProgress ( 100 );
    qApp->processEvents();
//     MyWait ( 1500 );
  }

//   qApp->processEvents();

  return true;
}

void QtDesignerWriter::recordCursorResource ( Resource* ) {
  qDebug ( "Sorry, not implemented: QtDesignerWriter::recordCursorResource" );
}


void QtDesignerWriter::recordBitmapResource ( Resource* ) {
  qDebug ( "Sorry, not implemented: QtDesignerWriter::recordBitmapResource" );
}


void QtDesignerWriter::recordIconResource ( Resource* ) {
  qDebug ( "Sorry, not implemented: QtDesignerWriter::recordIconResource" );
}


void QtDesignerWriter::recordMenuResource ( Resource* ) {
  qDebug ( "Sorry, not implemented: QtDesignerWriter::recordMenuResource()" );
}


void QtDesignerWriter::recordDialogResource ( QTextStream& str,
    Resource* current ) {
  nOurDlgFontSize = 12;          // needed for guessing height of ComboBoxes
  sOurDlgFontName = "helvetica"; // needed for breaking lines on QPushButtons

  nStorePlaceholder = 0;         // unknown user def. control class?

  // Opening tag
  str << "<widget class=\"QDialog\">" << endl;

  ResDialog* data = ( ResDialog* ) current->data();

  // Write name property
  str << "  <property name=\"name\">" << endl
  << "    <string>" << "Dialog_" << current->header()->name().toString() << "</string>" << endl
  << "  </property>" << endl;

  // Write geometry property
  str << "  <property name=\"geometry\">" << endl
  << "    <rect>" << endl
  << "      <x>" << scale ( data->header()->x ) << "</x>" << endl
  << "      <y>" << scale ( data->header()->y ) << "</y>" << endl
  << "      <width>" << scale ( data->header()->cx ) << "</width>" << endl
  << "      <height>" << scale ( data->header()->cy ) << "</height>" << endl
  << "    </rect>" << endl
  << "  </property>" << endl;

  str << "  <property name=\"caption\">" << endl
  << "    <string>" << expandSpecialChars ( data->header()->caption ) << "</string>" << endl
  << "  </property>" << endl;

  // Only write a font property if there was one in the original
  // resource file.
  if ( data->header()->wPointSize != 0 ) {

    /** In order to enable Designer to display special characters correctly
     *  we make use of a matching-table to replace each font-family name
     *  by an appropriate name making sense for Unix systems.
     *
     *  To compensate for different run-length we use a user-defined factor
     *  - the higher the value the larger the font to be used by us.
     *
     *  The user may define both, the font matching pairs and the factors by
     *  editing the file ".knutrc".
     *  In case the rc file in not located in the default directory and/or
     *  it has different name the user may specify a command line parameter
     *  like this:
     *
     *           --fonts:/home/NAME/.knutrc
     *
     *  By default this file is supposed to contain two lines:
     *
     *           [replace fonts]
     *           MS Shell Dlg:helvetica:0.8
     *
     *  Additional font-replacements may be defined on separate lines...
     *
     */
    float   nOurMult;
    if ( !GetFontMatch ( data->header()->fontName, sOurDlgFontName, nOurMult ) )
      nOurMult = 0.8;
    // store Font Size used in this dialog for later use
    // (e.g. Combo Boxes need this information)
    nOurDlgFontSize = ( int ) ( nOurMult * scale ( data->header()->wPointSize ) );

    str << "  <property name=\"font\">" << endl
    << "    <font>" << endl
    << "      <family>" << sOurDlgFontName << "</family>" << endl
    << "      <pointsize>"
    <<   nOurDlgFontSize
    << "</pointsize>" << endl
    << "    </font>" << endl
    << "  </property>" << endl;
  }

  // Here would be the right place to write more dialog properties

  // PENDING(kalle) Handle styles and extended styles.

  // Examine group ordering of controls
  findGroupScanClassnameGetBuddy ( data->controls(), 0 );
  // Now do the controls
  QListIterator<ControlData> it ( *data->controls() );
  ControlData* currentcd;
  while ( ( currentcd = it.current() ) ) {
    ++it;
    writeControl ( str, currentcd, data->controls(), 0 );
  }

  // closing tag
  str << "</widget>" << endl;
}


void QtDesignerWriter::findGroupScanClassnameGetBuddy ( void* _controls,
    ControlData* pGroup ) {
  QList<ControlData>* controls = ( QList<ControlData>* ) _controls;
  if ( controls ) {
    QListIterator<ControlData> it ( *controls );
    ControlData* currentcd;
    const Q_UINT16 x1 = pGroup ?      pGroup->x  : 0;
    const Q_UINT16 x2 = pGroup ? x1 + pGroup->cx : 0;
    const Q_UINT16 y1 = pGroup ?      pGroup->y  : 0;
    const Q_UINT16 y2 = pGroup ? y1 + pGroup->cy : 0;
    while ( ( currentcd = it.current() ) ) {
      ++it;
      /** adjust  currentcd->classId.id  and/or  currentcd->lStyle
       *  ========================================================
       *  /windows/Programme/Internet Explorer/Connection Wizard/icwoobe.exe
       *  contains a dialogue with a group box having class id -1 but style 7.
       *  (Obviously they use a special kind of "ICW_GROUPBOX"es there...)
       *
       *  The same is true for combo boxes, see dlg. "Anrufen" (in german version)
       *  of /windows/Programme/NetMeeting/conf.exe, they call it "ComboBoxEx32".
       *
       *  To solve this problem we do 2 things:
       *
       *  1. Look into resconv.rc file for [Replace Controls] section
       *     and see if there are name-pairs matching our situation.
       *
       *     Syntax of that section is like that:
       *
       *         [Replace Controls]
       *         BorShade:GROUPBOX
       *         HutschiKnurri:RADIOBUTTON
       *         MyGoodOldEdit:EDIT CENTER
       *
       *     and so on...
       *
       *     Valid names for the right sided term are:
       *
       *         BUTTON        CHECKBOX AUTOCHECKBOX 3STATEBOX
       *         AUTO3STATEBOX GROUPBOX RADIOBUTTON  AUTORADIOBUTTON
       *         EDIT          EDITLEFT EDITRIGHT    EDITCENTER
       *         STATIC        LISTBOX  SCROLLBAR    COMBOBOX
       *
       *  2. If user defined automatic matching did *not* work
       *     we try a more forgiving matching: look if certain
       *     patterns might be contained in the style name.
       *     There must be no exact match but the patterns are
       *     accepted even if combined with other letters.
       */
      QString sClassName;
      if ( !currentcd->classId.isordinal )
        sClassName = currentcd->classId.string;
      else
        if ( BUTTON == currentcd->classId.id )
          sClassName = currentcd->lStyle;
      if ( !sClassName.isEmpty() ) {
        // 1. Look if there is a user defined replacement
        //    for this user defined class in file resconv.rc
        if ( !GetCtrlMatch ( sClassName, currentcd->classId.id,
                             currentcd->lStyle ) ) {
          // 2. Try our own replacement, based of common patterns
          //    that might be *part*of* the class name: no exact
          //    matching is needed here.
          //
          // NOTE TO MY COLLEAGUES: If enhancing this list,
          //                        be sure to look for partial strings
          //                        AFTER looking for the whole string.
          //                  e.g.: first "AUTOCHECKBOX" then "CHECKBOX"
          //
          Q_UINT32 lStyleUpperBits = ( currentcd->lStyle & 0xFFFFFFF0 );
          if ( sClassName.contains ( "AUTOCHECKBOX",   false ) ) {
            currentcd->classId.id = BUTTON;
            currentcd->lStyle = lStyleUpperBits | BS_AUTOCHECKBOX;
          } else if ( sClassName.contains ( "CHECKBOX",       false ) ) {
            currentcd->classId.id = BUTTON;
            currentcd->lStyle = lStyleUpperBits | BS_CHECKBOX;
          } else if ( sClassName.contains ( "AUTORADIOBUTTON", false ) ) {
            currentcd->classId.id = BUTTON;
            currentcd->lStyle = lStyleUpperBits | BS_AUTORADIOBUTTON;
          } else if ( sClassName.contains ( "RADIOBUTTON",    false ) ) {
            currentcd->classId.id = BUTTON;
            currentcd->lStyle = lStyleUpperBits | BS_RADIOBUTTON;
          } else if ( sClassName.contains ( "AUTO3STATE",     false ) ) {
            currentcd->classId.id = BUTTON;
            currentcd->lStyle = lStyleUpperBits | BS_AUTO3STATE;
          } else if ( sClassName.contains ( "3STATE",         false ) ) {
            currentcd->classId.id = BUTTON;
            currentcd->lStyle = lStyleUpperBits | BS_3STATE;
          } else if ( sClassName.contains ( "GROUPBOX",       false ) ) {
            currentcd->classId.id = BUTTON;
            currentcd->lStyle = lStyleUpperBits | BS_GROUPBOX;
          }
          // In the following cases we don't have enough information
          // so we do NOT touch the Style here BUT ONLY set the Id.
          else if ( sClassName.contains ( "EDIT",           false ) )
            currentcd->classId.id = EDIT;
          else if ( sClassName.contains ( "STATIC",         false ) )
            currentcd->classId.id = STATIC;
          else if ( sClassName.contains ( "LISTBOX",        false ) )
            currentcd->classId.id = LISTBOX;
          else if ( sClassName.contains ( "SCROLLBAR",      false ) )
            currentcd->classId.id = SCROLLBAR;
          else if ( sClassName.contains ( "COMBOBOX",       false ) )
            currentcd->classId.id = COMBOBOX;
          // Now we process all remaining "Button"s...
          else if ( sClassName.contains ( "BUTTON",         false ) ) {
            currentcd->classId.id = BUTTON;
            if ( !BS_DEFPUSHBUTTON == ( currentcd->lStyle & 0x0000000FL ) )
              currentcd->lStyle = lStyleUpperBits | BS_PUSHBUTTON;
          }
          // And last but not least: Special greetings to all of you
          // out there using sophisticated Borland classes that might
          // contain the pattern "BorShade" in any combination.  :-)
          else if ( sClassName.contains ( "BORSHADE",       false ) ) {
            currentcd->classId.id = BUTTON;
            currentcd->lStyle = lStyleUpperBits | BS_GROUPBOX;
          }
        }
      }
      // Make sure ComboBoxes will be embedded into their Groups even if
      // they look larger in expanded state.
      Q_UINT16 cur_cy = currentcd->cy;
      // we are having raw (== unscaled) coordinate data here, so
      // we have to unscale the nOurDlgFontSize to use it propperly.
      Q_UINT16 nComboBoxHeight = unscale ( 11 + nOurDlgFontSize );
      if ( COMBOBOX ==  currentcd->classId.id
           && nComboBoxHeight < cur_cy ) {
        cur_cy = nComboBoxHeight;
      }
      if ( ( pGroup != currentcd )
           && ( !pGroup
                || ( x1   <= currentcd->x
                     && x2 + 1 >= currentcd->x + currentcd->cx
                     && y1   <= currentcd->y
                     && y2 + 1 >= currentcd->y + cur_cy ) ) ) {
        // Prevent pointer to inner group from being overwritten by outer
        // group pointer in case the outer one is declared later.
        bool bOk = true;
        if ( currentcd->myGroup ) {
          ControlData* pParent = currentcd;
//qDebug("\n");
          while ( pParent && bOk ) {
            if ( ( pParent->myGroup == pGroup ) ||
                 // detect not-allowed recursive grouping:
                 // (simple but fast code)
                 ( pParent->myGroup == currentcd ) ||
                 ( pParent->myGroup == pParent ) ||
                 ( pParent->myGroup &&
                   ( ( pParent->myGroup->myGroup == pParent
                     ) ||
                     ( pParent->myGroup->myGroup &&
                       ( ( pParent->myGroup->myGroup->myGroup == pParent
                         ) ||
                         ( pParent->myGroup->myGroup->myGroup &&
                           ( pParent->myGroup->myGroup->myGroup->myGroup == pParent
                           )
                         )
                       )
                     )
                   )
                 )
               )
              bOk = false;
//qDebug(QString("pParent->myGroup == pGroup   : %1").arg((pParent->myGroup == pGroup)   ?"true":"false"));
//qDebug(QString("pParent->myGroup == pParent  : %1").arg((pParent->myGroup == pParent)  ?"true":"false"));
//qDebug(QString("pParent->myGroup == currentcd: %1").arg((pParent->myGroup == currentcd)?"true":"false"));
//qDebug(QString("bOk: %1").arg(bOk?"true":"false"));
            pParent = pParent->myGroup;
          }
        }
        if ( bOk ) {
          currentcd->myGroup = pGroup;
          if ( ( BUTTON      == currentcd->classId.id )
               && ( BS_GROUPBOX == ( currentcd->lStyle & 0x0000000FL ) ) ) {
            findGroupScanClassnameGetBuddy ( controls, currentcd );
          }
        }
      }
    }
    // get Buddy for all controls with & in their name
    /*

    Sorry: This feature is still missing.

    if( !pGroup )
    {
      it.toFirst();
      while( (currentcd = it.current()) )
      {
        ++it;

      }
    }
    */
  }
}


void QtDesignerWriter::recordStringResource ( Resource* ) {
  qDebug ( "Sorry, not implemented: QtDesignerWriter::recordStringResource" );
}


void QtDesignerWriter::recordFontResource ( Resource* ) {
  qDebug ( "Sorry, not implemented: QtDesignerWriter::recordFontResource" );
}


void QtDesignerWriter::recordFontDirResource ( Resource* ) {
  qDebug ( "Sorry, not implemented: QtDesignerWriter::recordFontDirResource" );
}


void QtDesignerWriter::recordAcceleratorsResource ( Resource* ) {
  qDebug ( "Sorry, not implemented: QtDesignerWriter::recordAcceleratorsResource" );
}


void QtDesignerWriter::recordRCDataResource ( Resource* ) {
  qDebug ( "Sorry, not implemented: QtDesignerWriter::recordRCDataResource" );
}


void QtDesignerWriter::recordMessageTableResource ( Resource* ) {
  qDebug ( "Sorry, not implemented: QtDesignerWriter::recordMessageTableResource" );
}


void QtDesignerWriter::recordGroupCursorResource ( Resource* ) {
  qDebug ( "Sorry, not implemented: QtDesignerWriter::recordGroupCursorResource" );
}


void QtDesignerWriter::recordGroupIconResource ( Resource* ) {
  qDebug ( "Sorry, not implemented: QtDesignerWriter::recordGroupIconResource" );
}


void QtDesignerWriter::recordVersionResource ( Resource* ) {
  qDebug ( "Sorry, not implemented: QtDesignerWriter::recordVersionResource" );
}


void QtDesignerWriter::recordNewBitmapResource ( Resource* ) {
  qDebug ( "Sorry, not implemented: QtDesignerWriter::recordNewBitmapResource" );
}


void QtDesignerWriter::recordNewMenuResource ( Resource* ) {
  qDebug ( "Sorry, not implemented: QtDesignerWriter::recordNewMenuResource" );
}


void QtDesignerWriter::recordNewDialogResource ( Resource* ) {
  qDebug ( "Sorry, not implemented: QtDesignerWriter::recordNewDialogResource" );
}


void QtDesignerWriter::recordToolbarResource ( Resource* ) {
  qDebug ( "Sorry, not implemented: QtDesignerWriter::recordToolbarResource" );
}


void QtDesignerWriter::recordDlgInitResource ( Resource* ) {
  qDebug ( "Sorry, not implemented: QtDesignerWriter::recordDlgInitResource" );
}


void QtDesignerWriter::writeControl ( QTextStream& s,  ControlData* cd,
                                      void* _controls, ControlData* pGroup ) {
  // do nothing if group not matching
  if ( pGroup != cd->myGroup ) return;
  // Designer used local coordinates in groups - rustic MS format doesn't.
  if ( pGroup ) {
    if ( pGroup->x <= cd->x )
      cd->x -= pGroup->x;
    if ( pGroup->y <= cd->y )
      cd->y -= pGroup->y;
  }

  bool bStorePlaceholder = false;

  static int ctrlno = 0;

  ctrlno++;
  QString nostr = QString::number ( ctrlno );
  // PENDING(kalle) Support more styles
  if ( !cd->classId.isordinal ) {
    qDebug ( "Sorry. User-defined controls are ignored if they are" );
    qDebug ( "       not declared in the ~/.knutrc configuration file." );
  }

  // Note: Button Styles may *not* be combined with other Button Styles but
  //       Button styles may be combined with window styles (WS_*) by using
  //       the bitwise OR operator.
  //( see: http://support.microsoft.com/support/kb/articles/Q76/9/33.ASP )
  // Therefor we only look at the LSByte.  :-)
  const long nOurStyle       = cd->lStyle & 0x0000000FL;
  const bool bThisIsGroupBox = BS_GROUPBOX == nOurStyle;
  int nWidgetWidth  = scale ( cd->cx );
  int nWidgetHeight = scale ( cd->cy );
  int nWidgetY      = scale ( cd->y );

  switch ( cd->classId.id ) {
    case BUTTON:
      if ( BS_PUSHBUTTON    == nOurStyle ||
           BS_DEFPUSHBUTTON == nOurStyle ) {
        // push button
        s << "  <widget class=\"QPushButton\">" << endl;

        // calculate width of text to find out position(s) of line-break(s)
        // that we might have to insert
        const int nTextAreaWidth = nWidgetWidth - 7;
        QFont font ( sOurDlgFontName, nOurDlgFontSize );
        QFontMetrics fm ( font );
        QString sOurText ( cd->text.toString() );
        if ( 0 < sOurText.length() ) {
          const int nMinHeight = 11 + nOurDlgFontSize;
          if ( nWidgetHeight < nMinHeight ) {
            const int nDelta = nMinHeight - nWidgetHeight;
            if ( nWidgetY  > nDelta )
              nWidgetY -= nDelta;
            nWidgetHeight = nMinHeight;
          }
          if ( nTextAreaWidth
               < ( fm.size ( Qt::SingleLine | Qt::ShowPrefix, cd->text.toString()
                           ) ).width() ) {
            QString sLineBuf;
            uint nLastIns = 0;
            for ( uint iPos = 0; iPos < sOurText.length(); ++iPos ) {
              if ( nTextAreaWidth
                   < ( fm.size ( Qt::SingleLine | Qt::ShowPrefix, sLineBuf ) ).width() ) {
                bool bFound      = false;
                bool bBetterFound = false;
                uint nIns  = iPos;
                for ( uint i = iPos; i > 1 + nLastIns; --i ) {
                  if ( ' ' == sOurText.at ( i ) ) {
                    if ( !bFound ) {
                      nIns = i;
                      bFound = true;
                    }
                    QChar c = sOurText.at ( i - 1 );
                    if ( ( ( ',' == c ) || ( ';' == c ) || ( ')' == c ) ) && !bBetterFound ) {
                      nIns = i;
                      bBetterFound = true;
                    } else if ( '.' == c ) {
                      nIns = i;
                      break;
                    }
                  }
                }
                // insert line-break
                sOurText.insert ( nIns, "\x0A" );
                nLastIns = nIns;
                ++iPos;
                sLineBuf  = sOurText.at ( iPos ); // restart buffer
              } else
                sLineBuf += sOurText.at ( iPos ); // add to buffer
            }
          }
          sOurText = expandSpecialChars ( sOurText );
        }
        s << "    <property name=\"text\">" << endl
        << "      <string>" << sOurText << "</string>" << endl
        << "    </property>" << endl;
        if ( BS_DEFPUSHBUTTON == nOurStyle )
          s << "    <property name=\"default\">" << endl
          << "      <bool>true</bool>" << endl
          << "    </property>" << endl;
      } else if ( ( BS_CHECKBOX     == nOurStyle ) ||
                  ( BS_AUTOCHECKBOX == nOurStyle ) ||
                  ( BS_AUTO3STATE   == nOurStyle ) ||
                  ( BS_3STATE       == nOurStyle ) ) {
        s << "  <widget class=\"QCheckBox\">" << endl;
        s << "    <property name=\"text\">" << endl
        << "      <string>" << expandSpecialChars ( cd->text.toString() ) << "</string>" << endl
        << "    </property>" << endl;
      } else if ( bThisIsGroupBox ) {
        s << "  <widget class=\"QGroupBox\">" << endl;
        s << "    <property name=\"title\">" << endl
        << "      <string>" << expandSpecialChars ( cd->text.toString() ) << "</string>" << endl
        << "    </property>" << endl;
      } else if ( ( BS_RADIOBUTTON     == nOurStyle ) ||
                  ( BS_AUTORADIOBUTTON == nOurStyle ) ) {
        s << "  <widget class=\"QRadioButton\">" << endl;
        s << "    <property name=\"text\">" << endl
        << "      <string>" << expandSpecialChars ( cd->text.toString() ) << "</string>" << endl
        << "    </property>" << endl;
      }
      /* The following ones are included with their relatives above
         but *should* be processed separately 'some day'.
       else if( BS_AUTOCHECKBOX    == nOurStyle ||
                BS_AUTO3STATE      == nOurStyle ||
                BS_AUTORADIOBUTTON == nOurStyle ||
                //      BS_PUSHBOX == nOurStyle ||
                BS_3STATE          == nOurStyle )
        qDebug( "Sorry, unimplemented Button-like control" );
      */
      else
        bStorePlaceholder = true;
      break;
    case EDIT:
      s << "  <widget class=\"QLineEdit\">" << endl;
      // PENDING(kalle) How to distinguish between QLineEdit and QMultiLineEdit?
      break;
    case STATIC: {
      if ( SS_ICON == nOurStyle ) { // ICON resource
        s << "  <widget class=\"QLabel\">" << endl;
        s << "    <property name=\"pixmap\">" << endl
        << "      <pixmap>" << expandSpecialChars ( cd->text.toString() ) << "</pixmap>" << endl
        << "    </property>" << endl;
        SearchedIcon si;
        si.name = cd->text;
        si.cx = cd->cx;
        si.cy = cd->cy;
        imagesfordialog.append ( si );
        qDebug ( "Sorry, not implemented: Icons" );
      } else {
        // labels
        s << "  <widget class=\"QLabel\">" << endl;
        s << "    <property name=\"text\">" << endl
        << "      <string>" << expandSpecialChars ( cd->text.toString() ) << "</string>" << endl
        << "    </property>" << endl;
        if ( ES_RIGHT == nOurStyle ) // RTEXT resource
          s << "    <property name=\"alignment\">" << endl
          << "      <set>WordBreak|AlignRight|AlignVCenter</set>" << endl
          << "    </property>" << endl;
        else if ( ES_CENTER == nOurStyle ) // CTEXT resource
          s << "    <property name=\"alignment\">" << endl
          << "      <set>WordBreak|AlignCenter|AlignVCenter</set>" << endl
          << "    </property>" << endl;
        else // LTEXT resource
          s << "    <property name=\"alignment\">" << endl
          << "      <set>WordBreak|AlignLeft|AlignVCenter</set>" << endl
          << "    </property>" << endl;
        // Enlarge our widget a little bit (max. by the width of one '&')
        // to make it large enough for the text.
        QFont font ( sOurDlgFontName, nOurDlgFontSize );
        QFontMetrics fm ( font );
        int nTxtWidth = ( fm.size ( Qt::SingleLine,
                                    cd->text.toString() ) ).width();
        int nDeltaWidth = ( fm.size ( Qt::SingleLine, "&" ) ).width();
        if ( ( nTxtWidth - nDeltaWidth < nWidgetWidth )
             && ( nWidgetWidth < nTxtWidth ) )
          nWidgetWidth = nTxtWidth;
      }
      break;
    }
    case LISTBOX:
      s << "  <widget class=\"QListBox\">" << endl;
      // PENDING(kalle) More properties?
      break;
    case SCROLLBAR:
      s << "  <widget class=\"QScrollBar\">" << endl;
      // PENDING(kalle) More properties?
      break;
    case COMBOBOX: {
      s << "  <widget class=\"QComboBox\">" << endl;
      // problem: A Windows-ComboBox's height is not the height
      //          of the edit field but the entire height of the
      //          widget in expanded state.
      //          Unfortunately the not-expanded height is *not*
      //          stored separately but calculated based on the font used.
      //          So we have to guess the not-expanded height now
      //          and save it for usage below.
      int nGuessedHeight = nOurDlgFontSize + 11; // Ok for all font sizes
      if ( nWidgetHeight > nGuessedHeight )
        nWidgetHeight = nGuessedHeight;
      // PENDING(kalle) More properties?
      break;
    }
    default:
      bStorePlaceholder = true;
  }

  if ( bStorePlaceholder && ( 0 < nWidgetWidth ) && ( 0 < nWidgetHeight ) ) {
    ++nStorePlaceholder;
    const QString sNumber = QString::number ( nStorePlaceholder );
    s << "  <widget class=\"QLabel\">" << endl;
    s << "      <property name=\"pixmap\">" << endl;
    s << "          <pixmap>UserDefClass";
    s <<                    sNumber;
    s <<                    "</pixmap>" << endl;
    s << "      </property>" << endl;
    s << "      <property name=\"scaledContents\">" << endl;
    s << "          <bool>true</bool>" << endl;
    s << "      </property>" << endl;
    s << "      <property name=\"autoMask\">" << endl;
    s << "          <bool>true</bool>" << endl;
    s << "      </property>" << endl;
    // store information for usage in ::recordImagesForDialog() (see below)
    cd->nPlaceholder = nStorePlaceholder;
    cd->cx           = nWidgetWidth;
    cd->cy           = nWidgetHeight;
  }

  // all controls have a geometry
  s << "    <property name=\"geometry\">" << endl
  << "      <rect>" << endl
  << "        <x>"      << scale ( cd->x ) << "</x>" << endl
  << "        <y>"      << nWidgetY       << "</y>" << endl
  << "        <width>"  << nWidgetWidth   << "</width>" << endl
  << "        <height>" << nWidgetHeight  << "</height>" << endl
  << "      </rect>" << endl
  << "    </property>" << endl;

  s << "    <property name=\"name\">" << endl
  << "      <string>" << cd->wId << "</string>" << endl
  << "    </property>" << endl;

  // insert group childs now
  if ( bThisIsGroupBox ) {
    QList<ControlData>* controls = ( QList<ControlData>* ) _controls;
    if ( controls ) {
      QListIterator<ControlData> it ( *controls );
      ControlData* currentcd;
      while ( ( currentcd = it.current() ) ) {
        ++it;
        writeControl ( s, currentcd, controls, cd );
      }
    }
  }

  // footer for each control
  s << "  </widget>" << endl;
}


void paintPlaceholder ( QPainter *p, int w, int h, QString& sName ) {
  if ( ( 0 < w ) && ( 0 < h ) ) {
    p->eraseRect ( 0, 0, w, h );

    const int x1 = 1;
    const int y1 = 1;
    const int x2 = x1 + w - 2;
    const int y2 = y1 + h - 2;
    const int xBar = ( 12 <= x2 - x1 ) ? ( x2 - x1 ) / 4 : 5;
    const int yBar = ( 12 <= y2 - y1 ) ? ( y2 - y1 ) / 4 : 5;
    const int thick = 3;

    QPointArray a;
    QBrush brush ( Qt::yellow, Qt::SolidPattern );
    p->setBrush ( brush );

    a.setPoints ( 6, x1,      y1,
                  x1 + xBar, y1,
                  x1 + xBar, y1 + thick,
                  x1 + thick, y1 + thick,
                  x1 + thick, y1 + yBar,
                  x1,      y1 + yBar );
    p->drawPolygon ( a );
    a.setPoints ( 6, x2,      y1,
                  x2 - xBar, y1,
                  x2 - xBar, y1 + thick,
                  x2 - thick, y1 + thick,
                  x2 - thick, y1 + yBar,
                  x2,      y1 + yBar );
    p->drawPolygon ( a );
    a.setPoints ( 6, x1,      y2,
                  x1 + xBar, y2,
                  x1 + xBar, y2 - thick,
                  x1 + thick, y2 - thick,
                  x1 + thick, y2 - yBar,
                  x1,      y2 - yBar );
    p->drawPolygon ( a );
    a.setPoints ( 6, x2,      y2,
                  x2 - xBar, y2,
                  x2 - xBar, y2 - thick,
                  x2 - thick, y2 - thick,
                  x2 - thick, y2 - yBar,
                  x2,      y2 - yBar );
    p->drawPolygon ( a );

    QFont f ( "courier", 12, QFont::Normal );
    p->setFont ( f );
    /*
      QColor windowColor( 120, 120, 255 );  // a light blue color
      brush.setColor( windowColor );    // set this brush color
      p->setBrush( brush );     // set brush

      p->drawRect( 0, 0, 90, 50 );
    */
    p->drawText ( x1 + thick + 1, y1 + thick, x2 - thick, y2 - thick, Qt::AlignLeft, sName );
    /*
      p->setBackgroundMode( Qt::OpaqueMode );   // set opaque mode
      p->setBrush( Qt::DiagCrossPattern );    // black diagonal cross pattern
      p->drawEllipse( 90, 210, 80, 80 );    // back wheel
      p->setBrush( Qt::CrossPattern );    // black cross fill pattern
      p->drawEllipse( 310, 210, 80, 80 );   // front wheel
    */
  }
}


void storeImageData ( QTextStream& s, const QImage& image ) {
  // This is mostly from Qt Designer
  QByteArray ba;
  QBuffer buffer ( ba );
  buffer.open ( IO_WriteOnly );
  QImageIO imgio ( &buffer, "XPM" );
  imgio.setImage ( image );
  imgio.write();
  buffer.close();
  ulong len = ba.size() * 2;
  QByteArray bazip ( len );
  ::compress ( ( uchar* ) bazip.data(), &len, ( uchar* ) ba.data(), ba.size() );

  s << "    <data format=\"XPM.GZ\" length=\"" << ba.size() << "\">";
  static const char hexchars[] = "0123456789abcdef";
  for ( int i = 0; i < ( int ) len; ++i ) {
    uchar c = ( uchar ) bazip[i];
    s << hexchars[c >> 4];
    s << hexchars[c & 0x0f];
  }
  s << "</data>" << endl;
}


void QtDesignerWriter::recordImagesForDialog ( QTextStream& s,
    Resource* current ) {
  if ( ( 0 == imagesfordialog.count() ) && !nStorePlaceholder )
    return; // do not even write <images> ... </images>

  s << "<images>" << endl;

  if ( nStorePlaceholder && current ) {
    ResDialog* data = ( ResDialog* ) current->data();
    if ( data ) {
      QListIterator<ControlData> it ( *data->controls() );
      ControlData* currentcd;
      while ( ( currentcd = it.current() ) ) {
        ++it;
        if ( currentcd->nPlaceholder ) {
          QString sNumber = QString::number ( currentcd->nPlaceholder );
          s << "  <image name=\"UserDefClass"
          <<            sNumber
          <<            "\">" << endl;

          int w = currentcd->cx;
          int h = currentcd->cy;
          if ( 8 > w )
            w = 8;
          if ( 8 > h )
            h = 8;
          QString sName ( currentcd->classId.toString() );
          if ( h < 30 )
            sName.append ( ":" );
          else
            sName.append ( "\n" );
          sName.append ( currentcd->text.toString() );
          QPixmap pix ( w, h );
          QPainter paint;
          paint.begin ( &pix );
          paintPlaceholder ( &paint, w, h, sName );
          paint.end();
          pix.setMask ( pix.createHeuristicMask() );
          const QImage imgPlaceholder = pix.convertToImage();

          storeImageData ( s, imgPlaceholder );
          s << "  </image>" << endl;
        }
      }
    }
    nStorePlaceholder = 0;
  }

  for ( QValueList<SearchedIcon>::Iterator it = imagesfordialog.begin();
        it != imagesfordialog.end();
        ++it ) {
    s << "  <image name=\"" << expandSpecialChars ( ( *it ).name.toString() )
    << "\">" << endl;

    // Retrieve the image data.
    // First the icon group resource
    ResGroupIcon* icongroup = ResGroupIcon::findByNameOrOrdinal ( ( *it ).name );
    if ( !icongroup ) {
      s << "  </image>" << endl;
      break;
    }

    // In the icon group, find the icon that fits best.
    QString iconno = icongroup->findBestIcon ( ( *it ).cx, ( *it ).cy );
    // Now find the corresponding icon
    nameorordinal iconno1;
    iconno1.isordinal = false;
    iconno1.string = iconno;
    ResIcon* icon = ResIcon::findByNameOrOrdinal ( iconno1 );
    if ( !icon ) {
      s << "  </image>" << endl;
      break;
    }

    // Here is the image to write
    QImage image = icon->image();
    storeImageData ( s, image );
    s << "  </image>" << endl;
  }
  s << "</images>" << endl;
}


QString QtDesignerWriter::expandSpecialChars ( const QString& s ) {
  QString ret = s;
  ret = ret.replace ( QRegExp ( "&" ), "&amp;" );
  ret = ret.replace ( QRegExp ( "<" ), "&lt;" );
  ret = ret.replace ( QRegExp ( ">" ), "&gt;" );

  // khz: The following encoding is the very same used by Designer
  //      when storing .ui files containing special characters.
  ret = ret.replace ( QRegExp ( "�" ), "\xC3\xA4" );
  ret = ret.replace ( QRegExp ( "�" ), "\xC3\xB6" );
  ret = ret.replace ( QRegExp ( "�" ), "\xC3\xBC" );
  ret = ret.replace ( QRegExp ( "�" ), "\xC3\x9F" );
  ret = ret.replace ( QRegExp ( "�" ), "\xC3\x84" );
  ret = ret.replace ( QRegExp ( "�" ), "\xC3\x96" );
  ret = ret.replace ( QRegExp ( "�" ), "\xC3\x9C" );
  // this should be expanded...

  return ret;
}
