/* --------------------------------------------------------------------------
 * Copyright (C) 2000-2001 Klarlvdalens Datakonsult AB. All rights reserved.
 *
 * This file is part of KNUT, the Windows resource converter for Qt GUI
 * Designer. You may only use and modify it according to the license terms
 * appearing in the file LICENSE included in this package.
 *
 * Holders of a commercial license have additional rights as specified in the
 * license terms shipped with the commercial package.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * ------------------------------------------------------------------------ */


/****************************************************************************
**
** This class is the Resource File Writer part of the Res To X application.
** It is called by the main application.
** It calls internal class ResDirectory to transmit resources to .RES file.
**
*****************************************************************************/

#include "reswrite.h"

#include <qfile.h>
#include <qapplication.h>
#include <qdatastream.h>
#include <qtextstream.h>
#include <qmessagebox.h>
#include <qstring.h>
#include <iostream>

using namespace std;


#define MAGIC_EXE_ID 0x5A4D

#define PE_HD_ID 0x4550
#define NE_HD_ID 0x454E
#define SE_HD_ID 0x4552

// addresses relative to start of .EXE file
#define PTR_SEG_EXE_HD 0x0000003C

// addresses relative to start of Portable Exe Header
#define PTR_CNT_SECTIONS			0x00000006
#define PTR_SIZE_IMAGE_OPTIONAL_HD	0x00000014
#define OFS_IMAGE_OPTIONAL_HD		0x00000018


// addresses relative to IMAGE_OPTIONAL_HEADER
//#define PTR_IMAGE_BASE_PE       0x0000001C
//#define PTR_IMAGE_BASE_PE32PLUS 0x00000018
//#define PTR_RES_TAB_PE          0x00000070
//#define PTR_RES_TAB_PE32PLUS    0x00000080


/* Predefined resource types */
#define RT_NEWRESOURCE  0x2000
#define RT_ERROR        0x7fff
#define RT_CURSOR       0x0001
#define RT_BITMAP       0x0002
#define RT_ICON         0x0003
#define RT_MENU         0x0004
#define RT_DIALOG       0x0005
#define RT_STRING       0x0006
#define RT_FONTDIR      0x0007
#define RT_FONT         0x0008
#define RT_ACCELERATORS 0x0009
#define RT_RCDATA       0x000A
#define RT_MESSAGETABLE 0x000B
#define RT_GROUP_CURSOR 0x000C
#define RT_GROUP_ICON   0x000E
#define RT_VERSION      0x0010
#define RT_NEWBITMAP    (RT_BITMAP | RT_NEWRESOURCE)
#define RT_NEWMENU      (RT_MENU   | RT_NEWRESOURCE)
#define RT_NEWDIALOG    (RT_DIALOG | RT_NEWRESOURCE)
#define RT_DLGINIT      0x00F0 // MFC-specific
#define RT_TOOLBAR      0x00F1 // MFC-specific

/*****************************************************************************/
//  Resource Directory Entry in .EXE files with PE header
/*****************************************************************************/
static Q_UINT32 iEntryGlobal;
static Q_UINT32 iEntryGlobal2;

class ResDirectory;
class ResDirEntry
{
private:
	ResDirEntry*	pParent; // Directory we are contained in.

	Q_UINT32 nName;	// If MSB is set the remaining 31 bits point (rel. to the
					// start of resources) to an IMAGE_RESSOURCE_STRING_U
					// otherwise nName contains the Res-Id value.
	Q_UINT32 nData; // If MSB is set it points to a subdirectory (ResDirectory)
					// otherwise it points to the data (ResDataEntry).

	bool			bValid;
	bool			bStrName;
	bool			bIsDirectory;
	Q_UINT16*		pRawName;
	QString			sName; // needed for logging (when reading _or_ writing)
	Q_UINT16		nId;
	ResDirectory*	pDir;
	MyOut*			pOut;
	QDataStream*    pSrc;
	Q_UINT32		nOffsRTab;
	Q_UINT32		nRVARTab;
	QIODevice*		f;
	ResWriteResult* pResult;
    QProgressDialog* dlg;
private:
	bool MustRemoveActiveMSB(Q_UINT32& rValue)
	{
		if( 0x80000000 == (0x80000000 & rValue) )
		{
			rValue -= 0x80000000;
			return true;
		}
		else
			return false;
	}
	bool	Write_IdOrRawName( QDataStream& rRES );
	bool	IsUnknownResType();
	QString	TxtOfIdOrRawName(  char nMode, unsigned char nLen ) const;
	Q_UINT16 SizeOfIdOrRawName() // count of bytes to be written into .RES file
	{
		Q_UINT16 nSize = 1 + (pRawName ? *pRawName : 1);
		return 2*nSize;
	}
public:
	ResDirEntry() : pParent( 0 ), bValid( false ), pRawName( 0 ), nId( 0 ),
					pDir( 0 ), pSrc( 0 ), f( 0 ), pResult( 0 ), dlg( 0 )
	{}
	~ResDirEntry();
	bool Read_1_Raw(ResDirEntry*	 _pParent,
					MyOut&			 rOut,
					QDataStream&	 rSrc,
					Q_UINT32		 _nOffsRTab,
					Q_UINT32		 _nRVARTab,
					ResWriteResult&  rResult,
                    QProgressDialog* _dlg);
	bool Read_2_Name();		// read Name string into pRawName buffer
	bool Read_3_SubDir();	// read subdirectory
	bool Write_RES( QDataStream& rRES ); // output the resource into rRES
};



/*****************************************************************************/
//  Resource Directory in .EXE files with PE header
/*****************************************************************************/
class ResDirectory
{
	ResDirEntry*	 pParent; // 0 == Type, 1 == Name, 2 == Language
	MyOut&			 rOut;
	QDataStream&	 rSrc;
	Q_UINT32		 nOffsRTab;
	Q_UINT32		 nRVARTab;
	ResWriteResult&	 rResult;
    QProgressDialog* dlg;
	QIODevice*	 f;
	Q_UINT32	 nResFlags;
	Q_UINT32	 nDateTimeStamp;
	Q_UINT16	 nMajorVer;
	Q_UINT16	 nMinorVer;
	Q_UINT16	 nNamedEntries;
	Q_UINT16	 nIdEntries;
	ResDirEntry* pEntries;	// It contains nNamedEntries named entries
							// followed by nIdEntries entries with Ids.
public:
	ResDirectory( ResDirEntry*		_pParent,
				  MyOut&			_rOut,
				  QDataStream&		_rSrc,
				  Q_UINT32			_nOffsRTab,
				  Q_UINT32			_nRVARTab,
				  ResWriteResult&	_rResult,
                  QProgressDialog*  _dlg );
	~ResDirectory()
	{
		if( pEntries )
			delete[] pEntries;
	}
	Q_UINT16 GetEntries(){ return nNamedEntries + nIdEntries; }
	bool ReadEntries();
	bool Write_RES( QDataStream& rRES ); // output the resources into rRES
};



/*****************************************************************************/
//   Methods of class ResDirectory
/*****************************************************************************/
ResDirectory::ResDirectory( ResDirEntry*	_pParent,
							MyOut&			_rOut,
							QDataStream&	_rSrc,
							Q_UINT32		_nOffsRTab,
							Q_UINT32		_nRVARTab,
							ResWriteResult&	_rResult,
                            QProgressDialog* _dlg )
	: pParent( _pParent ), rOut( _rOut ), rSrc( _rSrc ),
	  nOffsRTab( _nOffsRTab ), nRVARTab( _nRVARTab ), rResult( _rResult ),
      dlg( _dlg )
{
	if( !rSrc.device() )  // internal program error!
	{
		rResult = RWInternalProgErr;
		return;
	}

	rResult = RWOk;

	f = rSrc.device();
	if( !f || IO_Ok != f->status() )
	{
		rResult = RWInternalProgErr; // NEVER call us without
		return;						 // an open and usable datasource
	}
	QString s = QString::number( f->at() ) + "%1:  Resource Directory found";
	rOut.AddLog( s);
	rSrc >> nResFlags;
	rSrc >> nDateTimeStamp;
	rSrc >> nMajorVer;
	rSrc >> nMinorVer;
	rSrc >> nNamedEntries;
	rSrc >> nIdEntries;
	s.sprintf(" with %d / %d entries (named/id)\n", nNamedEntries, nIdEntries);
	rOut.AddLog( s );
	if( IO_Ok != f->status() )
		rResult = RWSrcIOErr;
}

bool ResDirectory::ReadEntries()
{
	bool bOk  = true;
	pEntries  = new ResDirEntry[ nNamedEntries + nIdEntries ];
	Q_UINT16 iEntry;

    Q_UINT16 nDlgStep = dlg && (0 == iEntryGlobal) ? GetEntries() / 25 : 0;
    Q_UINT16 iDlgStep = 0;

	// to improve performance we read the resource information
	// in 3 steps (to avoid too much repositioning in the stream)

/*1*/
	for( iEntry = 0; iEntry < GetEntries(); ++iEntry, ++iEntryGlobal )
	{
        if( dlg )
        {
            QString s( "scanning dir entry #" );
            s += QString::number( iEntryGlobal );
            dlg->setLabelText( s );
            qApp->processEvents();
        }
		if( !pEntries[ iEntry ].Read_1_Raw( pParent, rOut, rSrc,
											nOffsRTab, nRVARTab, rResult,
                                            dlg ) )
		{
			bOk = false;	 // CRITICAL error
			break;
		}
	}
  if( dlg && !pParent )
  {
      QString s( "scanning dir entry #" );
      s += QString::number( iEntryGlobal );
      s += " OK";
      dlg->setLabelText( s );
      qApp->processEvents();
  }
/*2*/
	if( bOk )
    {
		for( iEntry = 0; iEntry < GetEntries(); ++iEntry )
		{
			if( !pEntries[ iEntry ].Read_2_Name() )
			{
				bOk = false; // CRITICAL error
				break;
			}
		}
  }
/*3*/
	if( bOk )
    {
		for( iEntry = 0; iEntry < GetEntries(); ++iEntry )
		{
			if( !pEntries[ iEntry ].Read_3_SubDir() )
			{
				bOk = false; // CRITICAL error
				break;
			}
            if( dlg && !pParent )
            {
                if( iDlgStep >= nDlgStep )
                {
                    dlg->setProgress( dlg->progress()+1 );
                    qApp->processEvents();
                    iDlgStep = 0;
                }
                iDlgStep++;
            }
		}
    }
	return bOk; // false means CRITICAL error while reading .EXE
}

bool ResDirectory::Write_RES( QDataStream& rRES ) // write resources into rRES
{
	bool bOk  = TRUE;
    Q_UINT16 nDlgStep = dlg ? iEntryGlobal / 25 : 0;
    Q_UINT16 iDlgStep = 0;
	for( Q_UINT16 iEntry = 0; iEntry < GetEntries(); ++iEntry, ++iEntryGlobal2 )
	{
        if( dlg )
        {
            QString s( "extracting resource #" );
            s += QString::number( iEntryGlobal2 );
            dlg->setLabelText( s );
            qApp->processEvents();
        }
		if( !pEntries[ iEntry ].Write_RES( rRES ) )
		{
			bOk = false;	 // CRITICAL error
			break;
		}
        if( dlg )
        {
            if( iDlgStep >= nDlgStep )
            {
                dlg->setProgress( dlg->progress()+1 );
                qApp->processEvents();
                iDlgStep = 0;
            }
            iDlgStep++;
        }
	}
  if( dlg && !pParent )
  {
      QString s( "extracting resource #" );
      s += QString::number( iEntryGlobal2 );
      s += " OK";
      dlg->setLabelText( s );
      qApp->processEvents();
  }
	return bOk;
}



/*****************************************************************************/
//  Methods of class ResDirEntry
/*****************************************************************************/
ResDirEntry::~ResDirEntry()
{
	if( pDir )
		delete pDir;
}

bool ResDirEntry::Read_1_Raw( ResDirEntry*	   _pParent,
							  MyOut&		   rOut,
							  QDataStream&	   rSrc,
							  Q_UINT32		   _nOffsRTab,
							  Q_UINT32		   _nRVARTab,
							  ResWriteResult&  rResult,
                              QProgressDialog* _dlg ) // only read raw data
{
	pParent		= _pParent,
	pOut		= &rOut;
	pSrc		= &rSrc;
	nOffsRTab	= _nOffsRTab;
	nRVARTab	= _nRVARTab;
	pResult		= &rResult;
    dlg         = _dlg;

	if( !pSrc->device() )  // internal program error!
	{
		*pResult = RWInternalProgErr;
		return false;
	}

	rResult = RWOk;
	f = pSrc->device();
	if( IO_Ok != f->status() )
	{
		*pResult = RWInternalProgErr; // NEVER call us without
		return false;				  // an open and usable datasource
	}

	*pSrc >> nName;
	bStrName = MustRemoveActiveMSB( nName );
	if( !bStrName )
		nId = (Q_UINT16)nName;

	*pSrc >> nData;
	bIsDirectory = MustRemoveActiveMSB( nData );
	if( IO_Ok != f->status() )
	{
		*pResult = RWSrcIOErr;
		return false;
	}

	bValid = true;
	return true;
}

bool ResDirEntry::Read_2_Name() // read Name string
{
	if( !pResult || !pSrc || !f || IO_Ok != f->status() )
	{
		*pResult = RWInternalProgErr; // NEVER call us without
		return false;				  // an open and usable datasource
	}
	bool bOk = true;
	if( bValid )
	{
		QString s;
		s.sprintf( "            id: %d", nId );
		if( bStrName )
		{
			if( !f->at( nOffsRTab + nName ) )
				bOk = false;
			else
			{
				Q_UINT16 nChar;
				// read and store the amount of chars to follow
				*pSrc >> nChar;
				pRawName = new Q_UINT16[ 1+nChar ];
				pRawName[ 0 ] = nChar;
				// read the chars
				Q_INT16 iChar = 1;
				do
				{
					*pSrc >>  pRawName[ iChar ];
					QChar qc( pRawName[ iChar ] );
					sName.append( qc );
					++iChar;
				} while( nChar >= iChar );
				if( IO_Ok != f->status() )
				{
					bOk = false;
					s.append( "  -- ERROR: COULD NOT READ RESSOURCE NAME --" );
				}
				else
				{
					s.append( ", name = \"" + sName + "\"" );
				}
			}
		}
		s.append( "\n" );
		pOut->AddLog( s );
	}
	return bOk;
}

bool ResDirEntry::Read_3_SubDir() // read Subdirectory if appropriate
{
	bool bOk = true;
	if( bValid && bIsDirectory && nData )
	{
		// find position of Subdirectory and read it
		if( !f->at( nOffsRTab + nData ) )
			bOk = false;
		else
		{
			pDir = new ResDirectory( this, *pOut, *pSrc,
									 nOffsRTab, nRVARTab, *pResult, dlg );
			if( pDir->GetEntries() )
				bOk = pDir->ReadEntries();
			else
				delete pDir;
				// empty directory are ignored, they are just strange, no error
		}
	}
	return bOk;
}


//static int nBmp = 0;// TEST007


bool ResDirEntry::Write_RES( QDataStream& rRES ) // write resource into rRES
{
	QIODevice* fr = rRES.device();
	if( !fr )
	{
		*pResult = RWInternalProgErr;
		return false;
	}

	if( pDir )
		return pDir->Write_RES( rRES );

	// retrieve TYPE, NAME and perhaps LANGUAGE from my parent ResDirEntries

	ResDirEntry* apDir[ 3 ];  //  0: TYPE  1: NAME  [2: LANGUAGE]
	apDir[ 0 ] = this;
	apDir[ 1 ] = 0;
	apDir[ 2 ] = 0;
	while( apDir[ 0 ]->pParent )
	{
		apDir[ 2 ] = apDir[ 1 ];
		apDir[ 1 ] = apDir[ 0 ];
		apDir[ 0 ] = apDir[ 0 ]->pParent;
	}


	if(     apDir[ 0 ]->IsUnknownResType() // is userdef type or zero or string
	    || !apDir[ 1 ]                     // no name found
//|| nBmp > 4// TEST007
	  )
	{
		return true; // Skip this resource
	}
//++nBmp;// TEST007
	Q_UINT16 nLanguage = apDir[ 2 ] ? apDir[ 2 ]->nId : 0xFFFF;

	// calculate size of header
	Q_UINT32 nSizeResHeader
		= 4									// DataSize
		+ 4									// HeaderSize
		+ apDir[ 0 ]->SizeOfIdOrRawName()	// TYPE
		+ apDir[ 1 ]->SizeOfIdOrRawName()	// NAME
		+ 4									// DataVersion
		+ 2									// MemoryFlags
		+ 2									// LanguageId
		+ 4									// Version
		+ 4;								// Characteristics

	// get data information from .EXE file
	*pResult = RWSrcIOErr;

	if( !f->at( nOffsRTab + nData ) )
		return false;

	Q_UINT32 nOffsResData,   nSizeResData,   nCodepage;
	*pSrc >> nOffsResData >> nSizeResData >> nCodepage;
	if( IO_Ok != f->status() )
		return false;

	// now get the data from .EXE file
	if(    !f->at( nOffsRTab - nRVARTab + nOffsResData )
		|| pSrc->atEnd() )
		return false;

	/* my buffer */ QByteArray aRawResData( nSizeResData );
	pSrc->readRawBytes( aRawResData.data(), nSizeResData );
	if( IO_Ok != f->status() )
		return false;

	// write the resource into the .RES file
	QString s;
	s = "    type: ";
	s += apDir[ 0 ]->TxtOfIdOrRawName( 0, 15 );
	s += "  name: ";
	s += apDir[ 1 ]->TxtOfIdOrRawName( 1, 14 );
	s += "language: ";
	bool bDefLang = ( 0xFFFF == nLanguage );	// No language stored?
	if(  bDefLang )
		nLanguage = 0x0407;
	QString s2;
	s2.setNum( nLanguage, 16 );
	s2 = s2.rightJustify( 4, '0' );
	if( !bDefLang )
		s += " 0x" + s2 + " ";
	else						// We _must_ have a language to know how to
	{							// convert characters from UNICODE to ANSI
		s += "(0x" + s2 + ")";	// so we use
	}							// the default language from .EXE header.
	s += "   size:%6d\n";
	s.sprintf( s, nSizeResData );
	pOut->AddLog( s );

	*pResult = RWDestIOErr;
	
	/* pad buffer */ Q_UINT32 aPadBytes = 0; // this are 4 * 0x00 bytes
	const char* pPad = (const char*)&aPadBytes;
	int nPad = 0;



	// padding needed before we start writing this resource?
  nPad = 4 - (fr->at() % 4);
  if( 4 > nPad )
    rRES.writeRawBytes( pPad, nPad );



	rRES << nSizeResData						// 4 DataSize
		 << nSizeResHeader;						// 4 HeaderSize
	if(		IO_Ok != fr->status()
		||	!apDir[0]->Write_IdOrRawName(rRES)	// x TYPE
		||	!apDir[1]->Write_IdOrRawName(rRES))	// y NAME
		return false;

	// padding needed?
  nPad = 4 - (fr->at() % 4);
  if( 4 > nPad )
    rRES.writeRawBytes( pPad, nPad );

	rRES << (Q_UINT32)0x00000000				// 4 DataVersion
		 << (Q_UINT16)0x1000					// 2 MemoryFlags (DISCARDABLE)
		 << nLanguage							// 2 LanguageId
		 << (Q_UINT32)0x00000000				// 4 Version
		 << (Q_UINT32)0x00000000;				// 4 Characteristics
	if( IO_Ok != fr->status() )
		return false;

	if( nSizeResData )
	{
		rRES.writeRawBytes( aRawResData.data(), nSizeResData );
	}

  nPad = 4 - (fr->at() % 4);
  if( 4 > nPad )
    rRES.writeRawBytes( pPad, nPad );
	if( IO_Ok != fr->status() )
		return false;

	*pResult = RWOk;
	return true;
}


bool ResDirEntry::Write_IdOrRawName( QDataStream& rRES )
{
	if( pRawName ) // String ?
	{
		Q_UINT16 nLen = *pRawName;
		if( nLen )
			rRES.writeRawBytes( ((char*)pRawName)+2, nLen );
		rRES << (Q_UINT16)0x0000;
	}
	else
	{
		rRES << (Q_UINT16)0xffff << nId;
	}
	return (IO_Ok == rRES.device()->status());
}


bool ResDirEntry::IsUnknownResType()
{
//return (nId != RT_BITMAP);// TEST007
	QString s = TxtOfIdOrRawName( 0,0 );
	return ("RT_" != s.left( 3 ));
}


QString ResDirEntry::TxtOfIdOrRawName( char nMode, unsigned char nLen ) const
{
	QString text;
	if( pRawName )
		text = sName;
	else
	{
		switch( nMode )
		{
		case 0: switch( nId ) // it's IMPORTANT that the names start with RT_
				{			  // because IsUnknownResType() looks for this
				case RT_CURSOR:
				  text += "RT_CURSOR";
				  break;
				case RT_BITMAP:
				  text += "RT_BITMAP";
				  break;
				case RT_ICON:
				  text += "RT_ICON";
				  break;
				case RT_MENU:
				  text += "RT_MENU";
				  break;
				case RT_DIALOG:
				  text += "RT_DIALOG";
				  break;
				case RT_STRING:
				  text += "RT_STRING";
				  break;
				case RT_FONTDIR:
				  text += "RT_FONTDIR";
				  break;
				case RT_FONT:
				  text += "RT_FONT";
				  break;
				case RT_ACCELERATORS:
				  text += "RT_ACCELERATORS";
				  break;
				case RT_RCDATA:
				  text += "RT_RCDATA";
				  break;
				case RT_MESSAGETABLE:
				  text += "RT_MESSAGETABLE";
				  break;
				case RT_GROUP_CURSOR:
				  text += "RT_GROUP_CURSOR";
				  break;
				case RT_GROUP_ICON:
				  text += "RT_GROUP_ICON";
				  break;
				case RT_VERSION:
				  text += "RT_VERSION";
				  break;
				case RT_NEWBITMAP:
				  text += "RT_NEWBITMAP";
				  break;
				case RT_NEWMENU:
				  text += "RT_NEWMENU";
				  break;
				case RT_NEWDIALOG:
				  text += "RT_NEWDIALOG";
				  break;
				case RT_DLGINIT:
				  text += "RT_DLGINIT";
				  break;
				case RT_TOOLBAR:
				  text += "RT_TOOLBAR";
				  break;
				default:
				  text.sprintf("0x%x", nId);
				  break;
				}
				break;
		case 1: text.sprintf( "%d", nId );
				break;
		}
	}
	if( nLen < text.length() )
		text = text.stripWhiteSpace();
	if( nLen > text.length() )
		text = text.leftJustify( nLen );
	return text;
}


/*****************************************************************************/
//  Methods of class ResWrite
/*****************************************************************************/
ResWrite::ResWrite( QWidget*         _parent,
                    QIODevice*       _exedata,
                    QIODevice*       _resdata,
                    MyOut&           _rOut,
                    ResWriteResult&  _rResult,
                    QProgressDialog* _dlg ) :
    parent( _parent ), rOut( _rOut ), rResult( _rResult ),
    dlg( _dlg )
{
    iEntryGlobal = 0;
    if(  !_exedata				 || !_resdata
	  || !_exedata->isOpen()	 || !_resdata->isOpen()
	  || !_exedata->isReadable() || !_resdata->isWritable() )
	{
		rResult = RWUserParameterErr;
		return;
	}

    QIODevice& f = *_exedata;

    rOut.SetStat( "NEW TASK: Extracting resources from file" );
    if( dlg )
    {
        dlg->setProgress( dlg->progress()+1 );
        qApp->processEvents();
    }

	QDataStream src(&f);
   	src.setByteOrder( QDataStream::LittleEndian );
	QString s;

	do
	{
		// check .EXE file signature
		Q_INT16 nId;
		src >> nId;
		if( SrcError( MAGIC_EXE_ID != nId,
					"File has no valid .EXE header signature!" ) )
		{
			rResult = RWSrcFormatErr;
			break;
		}

		// Ok, this seems to be an .EXE file, we start analyzing...
    	rOut.AddLog( "*** Analyzing source file\n\n" );

		// find Offset to Segmented Exe Header
		f.at( PTR_SEG_EXE_HD );
		if( SrcError( src.atEnd(),
						"Offset to Segmented Exe Header not found." ) )
			break;
		Q_UINT32 nOffsHd;
		src >> nOffsHd;
		if( SrcError( PTR_SEG_EXE_HD >= nOffsHd || f.size() <= nOffsHd+8,
						"Offset to Segmented Exe Header invalid.\n"
						"--> Might it be that this is a *compressed* file?" ) )
			break;
		// goto Segmented Exe Header and check signature
		f.at( nOffsHd );
		if( SrcError( src.atEnd(),
						"Offset to Segmented Exe Header not found." ) )
			break;
		Q_UINT16 nHdSig;
		src  >>  nHdSig;
		switch(  nHdSig )
		{
		case PE_HD_ID: // "PE" == Portable (extended) Executable Header
		{
			DumpPE( src, *_resdata, nOffsHd, rResult );
			break;
		}
		case NE_HD_ID: // "NE" == Extended Executable Header (old style)
		{
			DumpNE();
			break;	
		}
		case SE_HD_ID: // "SE" == (???) Executable Header
		{
			DumpSE();
			break;
		}
		default:
		{
 	    	s.sprintf(
				"\n\nSegmented Exe Header type not supported, signature: %x",
						nHdSig );
			SrcError( true, s );
			break;
		}
		}//switch	
	}
	while( 0 );

	src.unsetDevice();

	if( RWOk == rResult )
		rOut.SetStat("Resource file was written successfully.");
}


ResWrite::~ResWrite()
{
}


void ResWrite::DumpPE(	QDataStream&	rSrc,
						QIODevice&		fd,
						Q_UINT32		nOffsHd,
						ResWriteResult&	rResult )
{
	if(    !rSrc.device()
		|| !fd.isOpen()
		|| !fd.isWritable() )  // internal program error!
	{
		rResult = RWInternalProgErr;
		return;
	}

	rResult = RWSrcFormatErr;

	QIODevice& f = *rSrc.device();

	QString s;
	rOut.AddLog("    Source File has \"PE\" header signature.\n\n"
				"*** Looking for resources...\n\n" );
    if( dlg )
    {
        dlg->setLabelText( "PE header found" );
        dlg->setProgress( dlg->progress()+1 );
        qApp->processEvents();
    }

	rResult = RWOk;


	// find count of sections ( == number of entries in Section Table)
	Q_UINT32 nPtrCntSections = nOffsHd + PTR_CNT_SECTIONS;
	f.at( nPtrCntSections );
	if( SrcError( rSrc.atEnd(),
				"Number of Sections was not found." ) )
		return;
	Q_UINT16 nCntSections;
	rSrc >> nCntSections;
	if( SrcError( !nCntSections,
				"No resources! (Number of Sections is zero.)" ) )
		return;


	// find size of IMAGE_OPTIONAL_HEADER
	Q_UINT32 nPtrSizeImgOptHd = nOffsHd + PTR_SIZE_IMAGE_OPTIONAL_HD;
	f.at( nPtrSizeImgOptHd );
	if( SrcError( rSrc.atEnd(),
				"Size of IMAGE_OPTIONAL_HEADER not found." ) )
		return;
	Q_UINT16 nSizeImgOptHd;
	rSrc >> nSizeImgOptHd;
	if( SrcError( !nSizeImgOptHd,
				"No resources! (IMAGE_OPTIONAL_HEADER has zero length.)" ) )
		return;


	// goto IMAGE_OPTIONAL_HEADER and read magic word
	Q_UINT32 nPtrImgOptHd = nOffsHd + OFS_IMAGE_OPTIONAL_HD;
	f.at( nPtrImgOptHd );
	if( SrcError( rSrc.atEnd(),
				"No resources! (IMAGE_OPTIONAL_HEADER not found." ) )
		return;
	Q_UINT16 nIOHMagic;
	rSrc  >> nIOHMagic;

	bool bPE32Plus = false;
	switch( nIOHMagic )
	{
	case 0x010B: break;
	case 0x020B: bPE32Plus = true;
				 break;
	default: SrcError( true,
			"No resources! (Magic word of IMAGE_OPTIONAL_HEADER invalid.)" );
			return;
	}

	// goto Section Table and look for .rsrc entry
	f.at( nPtrImgOptHd + nSizeImgOptHd );
	if( SrcError( rSrc.atEnd(),
				"No resources! (Section Table not found.)" ) )
		return;
	char cSecName[8];
	Q_UINT32 nDummy;
	Q_UINT32 nRVARTab;
	Q_UINT32 nSizeRTab;
	Q_UINT32 nOffsRTab;
	Q_UINT32 nFlagsRTab;
	for( Q_UINT16 nSec = 0; nSec < nCntSections; ++nSec )
	{
		rSrc.readRawBytes( &cSecName[0], 8 );
		rSrc >> nDummy;		// virtual size
		rSrc >> nRVARTab;	// RVA  <--  we need it to find the resources' DATA
							//                      in ResDirEntry::Write_RES()
		rSrc >> nSizeRTab;	// physical size
		rSrc >> nOffsRTab;	// physical offset
		rSrc >> nDummy;		// reserved
		rSrc >> nDummy;		// reserved
		rSrc >> nDummy;		// reserved
		rSrc >> nFlagsRTab;	// section flags
		if( !strcmp(".rsrc", cSecName ) )
			break;
		nOffsRTab = 0;
	}
	if( SrcError( !nOffsRTab || !nSizeRTab,
				"No resource! (Res Table has zero offset or size.)" ) )
		return;
	if( f.size() < nOffsRTab+nSizeRTab )
	{
		QString s("No resource! (Offset to Resource Table invalid: ");
		s += QString::number( f.size() );
		s += ")";
		SrcError( TRUE,
					s + "--> Might it be that this is a *compressed* file?" );
		return;
	}
	// goto Resource Table and scan the root directory
	f.at( nOffsRTab );
	if( SrcError( rSrc.atEnd(),
				"No resource! (Offset to Resource Table is invalid)." ) )
		return;
	ResDirectory aRoot( 0, rOut, rSrc, nOffsRTab, nRVARTab, rResult, dlg );
	Q_UINT32 nEntries( aRoot.GetEntries() );
	if( SrcError( !nEntries || rSrc.atEnd(),
				"No resource entries are stored in this file." ) )
		return;

	// travel the root directory and all the Subdirectories
	// reading general info of each resource but NOT reading
	// the Resource Data
    if( dlg )
    {
        QString s( "root dir contains " );
        s += QString::number( nEntries );
        s += " entries";
        dlg->setLabelText( s );
        dlg->setProgress( dlg->progress()+1 );
        qApp->processEvents();
    }

    iEntryGlobal = 0;
	if( SrcError( !aRoot.ReadEntries(),
				"Conversion not possible. (Error reading .EXE file)." ) )
		return;


	QDataStream dest(&fd);
   	dest.setByteOrder( QDataStream::LittleEndian );

	rOut.AddLog( "\n\n*** Exporting resources to file\n" );

	// write dummy entry #0
	dest << (Q_UINT32)0x00000000	// DataSize      =  0
		 << (Q_UINT32)0x00000020	// HeaderSize    = 20

		 << (Q_UINT32)0x0000ffff	// Nm/Ord ResType = 0
		 << (Q_UINT32)0x0000ffff	//   "    ResName = 0

		 << (Q_UINT32)0x00000000	// DataVersion
		 << (Q_UINT16)0x0000		// MemoryFlags
		 << (Q_UINT16)0x0000		// LanguageId
		 << (Q_UINT32)0x00000000	// Version
		 << (Q_UINT32)0x00000000;	// Characteristics
	// write all the resources
    iEntryGlobal2 = 0;
	bool bOk = aRoot.Write_RES( dest );

	dest.unsetDevice();

	if( SrcError( !bOk,
				"Conversion not possible due to file i/o error." ) )
		return;

	rResult = RWOk;
}


void ResWrite::DumpNE()
{
    QString s;
   	rOut.AddLog("    Source File has \"NE\" header signature.\n\n"
"*** Sorry: Import of NE files is not implemented yet.\n\n" );
	rResult = RWSomeUnimplemented;


}


void ResWrite::DumpSE()
{
    QString s;
   	rOut.AddLog("    Source File has \"SE\" header signature.\n\n"
"*** Sorry: Import of SE files is not implemented yet.\n\n" );
	rResult = RWSomeUnimplemented;


}


QString ResWrite::GetResultTxt( ResWriteResult result )
{
	switch( result )
	{
	case RWOk:
			return QString("No errors, all resources converted successfully.");
	case RWUserParameterErr:
			return QString("Wrong parameters!\n\nUse: restox EXE-file-name RES-file-name [LOG-file-name].");
	case RWSomeUnimplemented:
			return QString("Not all of the resources were transformed (implementation missing).");
	case RWSomeUnknown:
			return QString("Not all of the resources were transformed (internal format unknown).");
	case RWSrcFormatErr:
			return QString("I encountered a syntax error in the .EXE file.");
	case RWSrcIOErr:
			return QString("Could not read from .exe file.");
	case RWDestIOErr:
			return QString("Could not write to .res file.");
	case RWLogIOErr:
			return QString("Could not write to .log file.");
	case RWInternalProgErr:
			return QString("An internal program error occured.");
	default:
			return QString("Unknown error, sorry.");
	}
}


bool ResWrite::SrcError( bool bError, QString sMsg )
{
	if( bError )
	{
    	QString s;
    	s.sprintf( "\n\nCRITICAL ERROR reading source file\n"
					"--> " + sMsg );
		if( parent )
			QMessageBox::information( parent, "Resource Dumper", s );
		rOut.AddLog( s + "\n\n" );
		rResult = RWSrcFormatErr;
	}
	return bError;
}

