/* --------------------------------------------------------------------------
 * Copyright (C) 2000-2001 Klar�lvdalens Datakonsult AB. All rights reserved.
 *
 * This file is part of KNUT, the Windows resource converter for Qt GUI
 * Designer. You may only use and modify it according to the license terms
 * appearing in the file LICENSE included in this package.
 *
 * Holders of a commercial license have additional rights as specified in the
 * license terms shipped with the commercial package.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * ------------------------------------------------------------------------ */


/**
 * $Id: resconv.cpp,v 1.8 2003/05/02 12:19:04 michel Exp $
 *
 * RES file reader.
 *
 * Implementation file for class ResConv.
 */

#include "resconv.h"

#include <myout.h>
#include <reswrite.h>
#include <ResFile.h>
#include <QtListViewWriter.h>
#include <QtDesignerWriter.h>
#include <QtSourceWriter.h>

#include <qbuffer.h>
#include <qcstring.h>
#include <qfile.h>

#include <stdio.h>

/*!
  This class provides a large number of static methods that serve as
  access points to the conversion functionalities in the resconv
  tool. Input formats can be Windows exe and dll files (methods
  starting with extract...) and Windows binary resource files (methods
  starting with convert...). Output formats can be Windows binary
  resource files (methods ending in ...Res), Qt Designer XML files
  (methods ending in ...XML) and Qt source (methods ending in
  ...Src). Output is also possible to a GUI display (methods ending in
  ...GUI).

  Each method comes in multiple incarnations. The input data can
  either be specified as a filename or a QIODevice, if the latter, it
  must already be opened for (at least) reading and be rewound to the
  beginning of the data. The output data can either be specified as a
  filename or a QIODevice, if the latter, it must already be opened
  for (at least) writing.

  If conversion fails or any of the files could not be opened, all the
  methods return true.

  This class should not be instantiated since all methods are
  private.
*/


/*!
  Extracts resource data from data representing a Windows exe or dll
  file specified as a QIODevice and writes it in Windows binary
  resource format to another QIODevice. The IO devices are not closed
  nor rewound after use.

  \param exedata The data representing the exe or dll file. The
  QIODevice must already be opened for (at least) reading and be
  rewound to the beginning.
  \param resdata The IO device to which the binary resource data
  should be written. It must already be opened for (at least)
  writing.
  \return true if the extraction succeeded and the data could be
  written to the output IO device, false otherwise.
*/
bool ResConv::extractToRes( QIODevice* exedata, QIODevice* resdata,
                            QProgressDialog *pdlg )
{
  MyOut aOut( 0, 0, 0, 0, false, true ); // bStdOut: 'false'   bStdErr: 'true'
  ResWriteResult result;
  ResWrite aRW( 0, exedata, resdata, aOut, result, pdlg );
  return ( result == RWOk );
}


/*!
  Extracts resource data from a Windows exe or dll file specified by
  its filename and writes it in Windows binary resource format to
  another file specified by its filename.

  \param exename The name of the Windows exe or dll file. The file
  must exist and be readable.
  \param resname The name of the file to write the binary resource
  data to. If the file already exists, it must be writable and will be
  truncated.
  \return true if the extraction succeeded and the data could be
  written to the output file, false otherwise.
*/
bool ResConv::extractToRes( const QString& exename, const QString& resname,
                            QProgressDialog *pdlg )
{
  bool success = false;
  QFile input( exename );
  if( input.open( IO_ReadOnly ) ) {
	QFile output( resname );
	if( output.open( IO_WriteOnly | IO_Truncate ) ) {
	  success = extractToRes( &input, &output, pdlg );
	  output.close();
	}
	input.close();
  }
  return success;
}


/*!
  Extracts resource data from a Windows exe or dll file specified by
  its filename and writes it in Windows binary resource format to a
  QIODevice. The output IO device will not be closed nor rewound after
  use.

  \param exename The name of the Windows exe or dll file. The file
  must exist and be readable.
  \param resdata The IO device to which the binary resource data
  should be written. It must already be opened for (at least)
  writing.
  \return true if the extraction succeeded and the data could be
  written to the output IO device, false otherwise.
*/
bool ResConv::extractToRes( const QString& exename, QIODevice* resdata,
                            QProgressDialog *pdlg )
{
  bool success = false;
  QFile input( exename );
  if( input.open( IO_ReadOnly ) ) {
	success = extractToRes( &input, resdata, pdlg );
	input.close();
  }
  return success;
}


/*!
  Extracts resource data from a Windows exe or dll file specified as a
  QIODevice, converts it to Qt Designer XML format and writes the resulting
  data to another IO device. The IO devices will not be closed nor rewound
  after use.

  \param exedata The data representing the exe or dll file. The
  QIODevice must already be opened for (at least) reading and be
  rewound to the beginning.
  \param xmldata The IO device to which the Qt Designer XML data should be
  written. It must already be opened for (at least) writing.
  \return true if the extraction succeeded and the data could be
  written to the output IO device, false otherwise.
*/
bool ResConv::extractToXML( QIODevice* exedata, QIODevice* xmldata,
                            float scalefactor, const QString& rcfile,
                            QProgressDialog *pdlg )
{
  // Use a buffer for storing the temporary res data.
  QBuffer resdata;
  if( !resdata.open( IO_ReadWrite ) )
    return false;

  bool success = false;
  // first step: extract from exe to res file
  if( extractToRes( exedata, &resdata, pdlg ) ) {
    resdata.reset(); // read from the beginning
    // second step: convert res file to xml
    success = convertToXML( &resdata, xmldata, scalefactor, rcfile,
                            pdlg );
  }
  resdata.close();
  return success;
}


/*!
  Extracts resource data from a Windows exe or dll file specified by its
  filename, converts it to Qt Designer XML format and writes the resulting
  data to another file specified by its filename.

  \param exename The name of the Windows exe or dll file. The file
  must exist and be readable.
  \param xmlname The name of the file to write the XML data to. If the file
  already exists, it must be writable and will be truncated.
  \return true if the extraction succeeded and the data could be
  written to the output file, false otherwise.
*/
bool ResConv::extractToXML( const QString& exename, const QString& xmlname,
                            float scalefactor, const QString& rcfile,
                            QProgressDialog *pdlg )
{
  bool success = false;

  QFile input( exename );
  if( input.open( IO_ReadOnly ) )
  {
	QFile output;
	bool bOk = TRUE;
	if( "-" == xmlname )
	  output.open( IO_WriteOnly, stdout );
	else
	{
	  output.setName( xmlname );
	  bOk = output.open( IO_WriteOnly | IO_Truncate );
	}
	if( bOk )
	{
	  success = extractToXML( &input, &output, scalefactor, rcfile,
                            pdlg );
	  output.close();
	}
	input.close();
  }
  return success;
}


/*!
  Extracts resource data from a Windows exe or dll file specified by its
  filename, converts it to Qt Designer XML format and writes the resulting
  data to a QIODevice. The QIODevice is not closed nor rewound after use.

  \param exename The name of the Windows exe or dll file. The file
  must exist and be readable.
  \param xmldata The IO device to which the Qt Designer XML data should be
  written. It must already be opened for (at least) writing.
  \return true if the extraction succeeded and the data could be
  written to the output IO device, false otherwise.
*/
bool ResConv::extractToXML( const QString& exename, QIODevice* xmldata,
                            float scalefactor, const QString& rcfile,
                            QProgressDialog *pdlg )
{
  bool success = false;
  QFile input( exename );
  if( input.open( IO_ReadOnly ) ) {
	success = extractToXML( &input, xmldata, scalefactor, rcfile, pdlg );
	input.close();
  }
  return success;
}


/*!
  Extracts resource data from a Windows exe or dll file specified as a
  QIODevice and displays a graphical representation in a listview. The IO
  device will not be closed nor rewound after use. The caller is responsible
  for layouting and showing the listview, as well as deleting it if no parent
  was specified.

  \param exedata The data representing the exe or dll file. The
  QIODevice must already be opened for (at least) reading and be
  rewound to the beginning.
  \param parent The parent widget for the listview which displays the resource
  data.
  \return If the extraction succeeded, a pointer to the listview widget is
  returned. If the extraction failed, 0 is returned.
*/
QListView* ResConv::extractToGUI( QIODevice* exedata, QWidget* parent )
{
  // Use a buffer for storing the temporary res data.
  QBuffer resdata;
  if( !resdata.open( IO_ReadWrite ) )
    return 0; //from: michel - Seems that Borland C++builder does not convert false to 0

  QListView* listview = 0;
  // first step: extract from exe to res file
  if( extractToRes( exedata, &resdata, 0 ) ) {
	resdata.reset(); // read from the beginning
	// second step: convert res file to xml
	listview = convertToGUI( &resdata, parent );
  }
  resdata.close();
  return listview;
}


/*!
  Extracts resource data from a Windows exe or dll file specified by its
  filename and displays a graphical representation in a listview. The caller
  is responsible for layouting and showing the listview, as well as deleting
  it if no parent was specified.

  \param exename The name of the exe or dll file.
  \param parent The parent widget for the listview which displays the resource
  data.
  \return If the extraction succeeded, a pointer to the listview widget is
  returned. If the extraction failed, 0 is returned.
*/
QListView* ResConv::extractToGUI( const QString& exename, QWidget* parent )
{
  QListView* listview = 0;
  QFile input( exename );
  if( input.open( IO_ReadOnly ) ) {
	listview = extractToGUI( &input, parent );
	input.close();
  }
  return listview;
}


/*!
  Extracts resource data from a Windows exe or dll file specified as a
  QIODevice, converts it to Qt source code and writes the resulting
  data to another IO device. The IO devices will not be closed nor rewound
  after use.

  \param exedata The data representing the exe or dll file. The
  QIODevice must already be opened for (at least) reading and be
  rewound to the beginning.
  \param srcdata The IO device to which the Qt source code should be
  written. It must already be opened for (at least) writing.
  \return true if the extraction succeeded and the data could be
  written to the output IO device, false otherwise.
*/
bool ResConv::extractToSrc( QIODevice* exedata, QIODevice* srcdata )
{
  // Use a buffer for storing the temporary res data.
  QBuffer resdata;
  if( !resdata.open( IO_ReadWrite ) )
	return false;

  bool success = false;
  // first step: extract from exe to res file
  if( extractToRes( exedata, &resdata, 0 ) ) {
	resdata.reset(); // read from the beginning
	// second step: convert res file to Qt source
	success = convertToSrc( &resdata, srcdata );
  }
  resdata.close();
  return success;
}


/*!
  Extracts resource data from a Windows exe or dll file specified by its
  filename, converts it to Qt source code and writes the resulting
  data to another file specified by its filename.

  \param exename The name of the Windows exe or dll file. The file
  must exist and be readable.
  \param srcname The name of the file to write the Qt source code data to. If
  the file already exists, it must be writable and will be truncated.
  \return true if the extraction succeeded and the data could be
  written to the output file, false otherwise.
*/
bool ResConv::extractToSrc( const QString& exename, const QString& srcname )
{
  bool success = false;
  QFile input( exename );
  if( input.open( IO_ReadOnly ) ) {
	QFile output( srcname );
	if( output.open( IO_WriteOnly | IO_Truncate ) ) {
	  success = extractToSrc( &input, &output );
	  output.close();
	}
	input.close();
  }
  return success;
}


/*!
  Extracts resource data from a Windows exe or dll file specified by its
  filename, converts it to Qt source code and writes the resulting
  data to a QIODevice. The QIODevice is not closed nor rewound after use.

  \param exename The name of the Windows exe or dll file. The file
  must exist and be readable.
  \param srcdata The IO device to which the Qt source code should be
  written. It must already be opened for (at least) writing.
  \return true if the extraction succeeded and the data could be
  written to the output IO device, false otherwise.
*/
bool ResConv::extractToSrc( const QString& exename, QIODevice* srcdata )
{
  bool success = false;
  QFile input( exename );
  if( input.open( IO_ReadOnly ) ) {
	success = extractToSrc( &input, srcdata );
	input.close();
  }
  return success;
}


/*!
  Converts Windows binary resource data from a res file specified as a
  QIODevice, converts it to Qt Designer XML format and writes the resulting
  data to another IO device. The IO devices will not be closed nor rewound
  after use.

  \param resdata The data representing the res file. The
  QIODevice must already be opened for (at least) reading and be
  rewound to the beginning.
  \param xmldata The IO device to which the Qt Designer XML data should be
  written. It must already be opened for (at least) writing.
  \return true if the extraction succeeded and the data could be
  written to the output IO device, false otherwise.
*/
bool ResConv::convertToXML( QIODevice* resdata, QIODevice* xmldata,
                            float scalefactor, const QString& rcfile,
                            QProgressDialog *pdlg )
{
  bool success = false;
  ResFile file( resdata );
  if( file.read() ) {
	QtDesignerWriter writer( xmldata, scalefactor, rcfile, pdlg );
	success = writer.write( file.resources() );
  }
  qDebug("ResConv::convertToXML");
  return success;
}


/*!
  Converts Windows binary resource data from a res file specified by its
  filename, converts it to Qt Designer XML format and writes the resulting
  data to another file specified by its filename.

  \param resname The name of the Windows res file. The file
  must exist and be readable.
  \param xmlname The name of the file to write the XML data to. If the file
  already exists, it must be writable and will be truncated.
  \return true if the extraction succeeded and the data could be
  written to the output file, false otherwise.
*/
bool ResConv::convertToXML( const QString& resname, const QString& xmlname,
                            float scalefactor, const QString& rcfile,
                            QProgressDialog *pdlg )
{
  bool success = false;
  QFile input( resname );
  if( input.open( IO_ReadOnly ) ) {
	QFile output( xmlname );
	if( output.open( IO_WriteOnly | IO_Truncate ) ) {
	  success = convertToXML( &input, &output, scalefactor, rcfile, pdlg );
	  output.close();
	}
	input.close();
  }
  return success;
}


/*!
  Converts Windows binary resource data from a Windows res file specified by
  its filename, converts it to Qt Designer XML format and writes the resulting
  data to a QIODevice. The QIODevice is not closed nor rewound after use.

  \param resname The name of the Windows res file. The file
  must exist and be readable.
  \param xmldata The IO device to which the Qt Designer XML data should be
  written. It must already be opened for (at least) writing.
  \return true if the extraction succeeded and the data could be
  written to the output IO device, false otherwise.
*/
bool ResConv::convertToXML( const QString& resname, QIODevice* xmldata,
                            float scalefactor, const QString& rcfile,
                            QProgressDialog *pdlg )
{
  bool success = false;
  QFile input( resname );
  if( input.open( IO_ReadOnly ) ) {
	success = convertToXML( &input, xmldata, scalefactor, rcfile, pdlg );
	input.close();
  }
  return success;
}


/*!
  Extracts resource data from a Windows res file specified as a
  QIODevice and displays a graphical representation in a listview. The IO
  device will not be closed nor rewound after use. The caller is responsible
  for layouting and showing the listview, as well as deleting it if no parent
  was specified.

  \param resdata The data representing the res file. The QIODevice must
  already be opened for (at least) reading and be rewound to the beginning.
  \param parent The parent widget for the listview which displays the resource
  data.
  \return If the extraction succeeded, a pointer to the listview widget is
  returned. If the extraction failed, 0 is returned.
*/
QListView* ResConv::convertToGUI( QIODevice* resdata, QWidget* parent )
{
  QListView* listview = 0;
  ResFile file( resdata );
  if( file.read() ) {
	QtListViewWriter writer( parent );
	listview = writer.write( file.resources() );
  }
  return listview;
}


/*!
  Extracts resource data from a Windows res file specified by its
  filename and displays a graphical representation in a listview. The caller
  is responsible for layouting and showing the listview, as well as deleting
  it if no parent was specified.

  \param resname The name of the res file.
  \param parent The parent widget for the listview which displays the resource
  data.
  \return If the extraction succeeded, a pointer to the listview widget is
  returned. If the extraction failed, 0 is returned.
*/
QListView* ResConv::convertToGUI( const QString& resname, QWidget* parent )
{
  QListView* listview = 0;
  QFile input( resname );
  if( input.open( IO_ReadOnly ) ) {
	listview = convertToGUI( &input, parent );
	input.close();
  }
  return listview;
}


/*!
  Converts Windows binary resource data from a res file specified as a
  QIODevice, converts it to Qt source code and writes the resulting
  data to another IO device. The IO devices will not be closed nor rewound
  after use.

  \param resdata The data representing the res file. The
  QIODevice must already be opened for (at least) reading and be
  rewound to the beginning.
  \param srcdata The IO device to which the Qt source code should be
  written. It must already be opened for (at least) writing.
  \return true if the extraction succeeded and the data could be
  written to the output IO device, false otherwise.
*/
bool ResConv::convertToSrc( QIODevice* resdata, QIODevice* srcdata )
{
  bool success = false;
  ResFile file( resdata );
  if( file.read() ) {
	QtSourceWriter writer( srcdata );
	success = writer.write( file.resources() );
  }
  return success;
}


/*!
  Converts Windows binary resource data from a res file specified by its
  filename, converts it to Qt source code and writes the resulting
  data to another file specified by its filename.

  \param resname The name of the Windows res file. The file
  must exist and be readable.
  \param xmlname The name of the file to write the Qt source code to. If the
  file already exists, it must be writable and will be truncated.
  \return true if the extraction succeeded and the data could be
  written to the output file, false otherwise.
*/
bool ResConv::convertToSrc( const QString& resname, const QString& srcname )
{
  bool success = false;
  QFile input( resname );
  if( input.open( IO_ReadOnly ) ) {
	QFile output( srcname );
	if( output.open( IO_WriteOnly | IO_Truncate ) ) {
	  success = convertToSrc( &input, &output );
	  output.close();
	}
	input.close();
  }
  return success;
}


/*!
  Converts Windows binary resource data from a Windows res file specified by
  its filename, converts it to Qt source code and writes the resulting
  data to a QIODevice. The QIODevice is not closed nor rewound after use.

  \param resname The name of the Windows res file. The file
  must exist and be readable.
  \param xmldata The IO device to which the Qt source code should be
  written. It must already be opened for (at least) writing.
  \return true if the extraction succeeded and the data could be
  written to the output IO device, false otherwise.
*/
bool ResConv::convertToSrc( const QString& resname, QIODevice* srcdata )
{
  bool success = false;
  QFile input( resname );
  if( input.open( IO_ReadOnly ) ) {
	success = convertToSrc( &input, srcdata );
	input.close();
  }
  return success;
}


bool extractToXML( const QString& exename, const QString& xmlname,
                   float scalefactor, const QString& rcfile )
{
  return ResConv::extractToXML( exename, xmlname, scalefactor, rcfile, 0 );
}


bool convertToXML( const QString& resname, const QString& xmlname,
                   float scalefactor, const QString& rcfile )
{
  return ResConv::convertToXML( resname, xmlname, scalefactor, rcfile, 0 );
}

