/* --------------------------------------------------------------------------
 * Copyright (C) 2000-2001 Klarlvdalens Datakonsult AB. All rights reserved.
 *
 * This file is part of KNUT, the Windows resource converter for Qt GUI
 * Designer. You may only use and modify it according to the license terms
 * appearing in the file LICENSE included in this package.
 *
 * Holders of a commercial license have additional rights as specified in the
 * license terms shipped with the commercial package.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * ------------------------------------------------------------------------ */


/**
 * RES file reader.
 *
 * Implementation file for class ResBitmap.
 */

#include "ResBitmap.h"



/***************************************************************
   fake bmp-header for resource import (code from qimage.cpp)
 ***************************************************************/

const int BMP_FILEHDR_SIZE = 14;	// size of BMP_FILEHDR data

struct BMP_FILEHDR			// BMP file header
{
    char   bfType[2];		// "BM"
    Q_INT32  bfSize;		// filesize ignored by qimage.cpp::read_bmp_image()
    Q_INT16  bfReserved1;
    Q_INT16  bfReserved2;
    Q_INT32  bfOffBits;		// pointer to the pixmap bits(also ignored)
	BMP_FILEHDR()
	{
		bfType[0]	= 'B';
		bfType[1]	= 'M';
		bfSize		= 0;
		bfReserved1	= 0;
		bfReserved2	= 0;
		bfOffBits	= 0;
	}
};

QDataStream &operator<<( QDataStream &s, const BMP_FILEHDR &bf )
{						// write file header
    s.writeRawBytes( bf.bfType, 2 );
    s << bf.bfSize << bf.bfReserved1 << bf.bfReserved2 << bf.bfOffBits;
    return s;
}






QString ResBitmap::GetCompressionTxt() const
{
	switch( ms.Compression )
	{
	case BCA_UNCOMP:
			return QString("none");
			break;
	case BCA_RLE8:
			return QString("RLE8");
			break;
	case BCA_RLE4:
			return QString("RLE4");
			break;
	case BCA_HUFFMAN1D:
			return QString("HUFFMAN1D");
			break;
	case BCA_RLE24:
			return QString("RLE24");
			break;
	default:
			return QString("unknown");
			break;
	}
}


void ResBitmap::read_DIB( Q_UINT16 nBFT, QDataStream& s )
{
  _restype = nBFT;

  QIODevice& rS = *s.device();
  int nStartPos = rS.at();

  bool bIsBitmap = BFT_BMAP == _restype;

  if( bIsBitmap || BFT_ICON == _restype )
  {
    bool bWindows;
    Q_UINT32 nColorsTotal;


    // 1st read the header

    s >>      ms.SizeHd;
    if( 12 == ms.SizeHd )	// OS/2 1.x uncompressed bitmap
    {
      bWindows = false;
      Q_UINT16 w;
      s >> w;  ms.Width   = w;
      s >> w;  ms.Height  = w;
	  s >>     ms.Planes;
      s >>     ms.BitCount;
               ms.Compression = BCA_UNCOMP;

      nColorsTotal = ((unsigned long)1 << ms.BitCount);
    }
    else
    {
      bWindows = true;
      s >> ms.Width;
      s >> ms.Height;
	  s >> ms.Planes;
      s >> ms.BitCount;

      nColorsTotal = ((unsigned long)1 << ms.BitCount);

	  if( 16 < ms.SizeHd )  s >> ms.Compression;
      else                         ms.Compression = BCA_UNCOMP;

	  if( 20 < ms.SizeHd )  s >> ms.SizeImage;
	  if( 24 < ms.SizeHd )  s >> ms.XPelsPerMeter;
	  if( 28 < ms.SizeHd )  s >> ms.YPelsPerMeter;

	  if( 32 < ms.SizeHd )
      {
        s >> ms.ColorsUsed;
        if( 24 != ms.BitCount  && !ms.ColorsUsed )
          ms.ColorsUsed = nColorsTotal;
        else if( ms.ColorsUsed > nColorsTotal ) // NOT trust ColorsUsed!
          ms.ColorsUsed = nColorsTotal;
      }
      else
        ms.ColorsUsed = nColorsTotal;

	  if( 36 < ms.SizeHd )  s >> ms.ColorsImportant;

	  if( 40 < ms.SizeHd )  s >> ms.Units;
	  if( 42 < ms.SizeHd )  s >> ms.Reserved;
	  if( 44 < ms.SizeHd )  s >> ms.Recording;
	  if( 46 < ms.SizeHd )  s >> ms.Rendering;
	  if( 48 < ms.SizeHd )  s >> ms.Size1;
	  if( 52 < ms.SizeHd )  s >> ms.Size2;
	  if( 56 < ms.SizeHd )  s >> ms.ColorEncoding;
	  if( 60 < ms.SizeHd )  s >> ms.Identifier;
    }

    QString sformat = bIsBitmap ? "bmp" : "icon";
    sformat += " size = %d, bmp dimensions = %d/%d, bit planes = %d, bits per pixel = %d, compression = ";
    sformat += GetCompressionTxt();
    qDebug( sformat,
            ms.SizeImage,
            ms.Width,
            ms.Height,
            ms.Planes,
            ms.BitCount );

    if( bIsBitmap )
    {
      // 2nd store fake bmp-header into the stream

      char aBuffer[ BMP_FILEHDR_SIZE ];
      if( BMP_FILEHDR_SIZE > nStartPos )
      {
        qDebug( "\nERROR: cannot import bitmap - no space to write fake-header\n" );
        return;
      }
      else
      {
        rS.at( nStartPos - BMP_FILEHDR_SIZE );
        s.readRawBytes( aBuffer, BMP_FILEHDR_SIZE );
        rS.at( nStartPos - BMP_FILEHDR_SIZE );
        BMP_FILEHDR aFileHdr;
        s << aFileHdr;
        rS.at( nStartPos - BMP_FILEHDR_SIZE );
      }


      // 3rd let Qt do the job for us: read colour palette and image data

      QImageIO iio;
      iio.setIODevice( &rS );
      iio.setFormat( "BMP" );
      if( iio.read() )
        _image = iio.image();
      else
        qDebug( "\nERROR: QImageIO could not read this bitmap\n" );
      if( _image.isNull() )
        qDebug( "\nERROR: QImageIO imported an empty image\n" );


      // 4th restore original content of the stream

      rS.at( nStartPos - BMP_FILEHDR_SIZE );
      s.writeRawBytes( aBuffer, BMP_FILEHDR_SIZE );
      rS.at( nStartPos - BMP_FILEHDR_SIZE );
    }
    else // so it is *not* a Bitmap, but an ICON
    {

      // 2nd read the palette


      /** Note: the height field represents both the XOR and AND masks.
          Before reading the data the height has to be divided by two.
        */
      ms.Height /= 2;

      Q_UINT16 ourBitCount = 4 == ms.BitCount ? 8 : ms.BitCount;

      if( 24 != ms.BitCount )
      {
        _image.create( ms.Width, ms.Height,
                       ourBitCount,
                       nColorsTotal,
                       1 == ms.BitCount ? QImage::BigEndian : QImage::IgnoreEndian );
        unsigned int i;
        unsigned char nR;
        unsigned char nG;
        unsigned char nB;
        for ( i=0; i < nColorsTotal; ++i ) // build color table
        {
          s >> nB;
          s >> nG;
          s >> nR;
          _image.setColor( i, qRgb( nR, nG, nB ) );

          if( bWindows )
            s >> nB; // ignore 4th byte of each mswin-palette entry
        }
      }
      else
        _image.create( ms.Width, ms.Height, 32,
                       nColorsTotal, QImage::IgnoreEndian );


      // 3rd read the RGB values ( == the XOR data )
      // using code from qt/kernel/qimage.cpp to read uncompressed bitmap data

      int w = ms.Width, h = ms.Height;
      int bpl = _image.bytesPerLine();
      uchar **line = _image.jumpTable();

      if ( ms.BitCount == 1 )				// 1 bit ICON image
	  {
        while ( --h >= 0 )
        {
	      if ( rS.readBlock((char*)line[h],bpl) != bpl )
            break;
        }
      }
      else if ( ms.BitCount == 4 )			// 4 bit ICON image
      {
        int  buflen = ((w+7)/8)*4;
        uchar *buf  = new uchar[buflen];
	    while ( --h >= 0 )
        {
		  if ( rS.readBlock((char*)buf,buflen) != buflen )
		    break;
		  register uchar *p = line[h];
		  uchar *b = buf;
		  for ( int i=0; i<w/2; i++ )	// convert nibbles to bytes
          {
		    *p++ = *b >> 4;
		    *p++ = *b++ & 0x0f;
  		  }
		  if ( w & 1 )			// the last nibble
		    *p = *b >> 4;
	    }
	    delete [] buf;
      }
      else if ( ms.BitCount == 8 )			// 8 bit ICON image
      {
	    while ( --h >= 0 )
        {
  	      if ( rS.readBlock((char *)line[h],bpl) != bpl )
		    break;
        }
      }
      else if ( ms.BitCount == 24 )			// 24 bit ICON image
      {
        register QRgb *p;
        QRgb  *end;
        uchar *buf24 = new uchar[bpl];
        int    bpl24 = ((w*24+31)/32)*4;
        uchar *b;
        while ( --h >= 0 )
        {
          p = (QRgb *)line[h];
          end = p + w;
          if ( rS.readBlock( (char *)buf24,bpl24) != bpl24 )
            break;
          b = buf24;
          while ( p < end )
          {
            *p++ = qRgb(b[2],b[1],b[0]);
            b += 3;
          }
        }
        delete[] buf24;
      }


      // 4th read the alpha mask ( == the AND data )
      // and merge it into the image

      bool bMustConvert = 24 != ms.BitCount;
      if( bMustConvert )
        _image = _image.convertDepth( 32 );  // temporarily set colour depth to 32 bit

      _image.setAlphaBuffer( true );

      UINT32* p;
      int x;
      w = ms.Width;
      h = ms.Height;
      UINT32 nPad = 32 - (w % 32);
      if( 32 > nPad )
        nPad /= 8;
      else
        nPad = 0;
      bpl = (w+7)/8;
      uchar* mask = new uchar[bpl];
      line = _image.jumpTable();
      while( --h >= 0 )
      {
        if ( rS.readBlock((char*)mask, bpl) != bpl )
          break;
        rS.at( rS.at() + nPad );
        p = (UINT32 *)line[h];
        for( x=0; x < w; ++x, ++p )
        {
          if( (0x01 << (7-x%8)) & mask[x/8] )
            *p &= 0x00ffffff; // 0x00nnnnnn == transparent
          else
            *p |= 0xff000000; // 0xffnnnnnn == opaque
        }
      }

      if( bMustConvert )
		_image = _image.convertDepth( ourBitCount );

      delete[] mask;
    }

    rS.at( nStartPos - BMP_FILEHDR_SIZE );
  }
}


QString ResBitmap::toString() const
{
  QString sformat( "\tresource data:\n\t\tsize = %d\n\t\t" );
  switch( _restype )
  {
    case BFT_BMAP:
      sformat += "bmp";
      break;
    case BFT_ICON:
      sformat += "icon";
      break;
    default:
      sformat += "UNKNOWN BITMAP RESSOURCE TYPE\n\t\t";
      break;
  }
  sformat += " dimensions = %d/%d\n\t\tbit planes = %d\n\t\tbits per pixel = %d\n\t\tcompression = ";
  QString text;
  text.sprintf( sformat,
                ms.SizeImage,
                ms.Width,
                ms.Height,
                ms.Planes,
                ms.BitCount );
  text += GetCompressionTxt();

  return text;
}


QDataStream& operator>>( QDataStream& s, ResBitmap& r )
{
  r.read_DIB( BFT_BMAP, s );
  return s;
}

