/* --------------------------------------------------------------------------
 * Copyright (C) 2000-2001 Klarlvdalens Datakonsult AB. All rights reserved.
 *
 * This file is part of KNUT, the Windows resource converter for Qt GUI
 * Designer. You may only use and modify it according to the license terms
 * appearing in the file LICENSE included in this package.
 *
 * Holders of a commercial license have additional rights as specified in the
 * license terms shipped with the commercial package.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * ------------------------------------------------------------------------ */


/**
 * RES file reader.
 *
 * Implementation file for class ResDialog.
 */

#include <qstring.h>
#include "ResDialog.h"
#include "typedefs.h"
#include "tools.h"

QString ResDialog::toString() const
{
  QString text;
  text += "\tresource data:\n";
  text += "\t\tnumber of items: " + QString::number( (int)dlgheader.numberOfItems ) + '\n';
  text += "\t\tcaption: " + dlgheader.caption + '\n';
  QListIterator<ControlData> it( _controls );
  ControlData* current;
  uint count = 0;
  while( ( current = it.current() ) ) {
	text += QString( "\t\tdialog control #" ) + QString::number( count ) + '\n';
	text += QString( "\t\t\tclass id = " ) + current->classId.toString() + '\n';
	text += QString( "\t\t\ttext = " ) + current->text.toString() + '\n';
	count++;
	++it;
  }

  return text;
}



void SkipThePadding( QDataStream& s )
{
  int pos = s.device()->at();
  uchar c;
  if( pos % 4 )
  {
    int skip = 4 - ( pos % 4 );
    for( int i = 0; i < skip; ++i )
 	  s >> c;
  }
}



QDataStream& operator>>( QDataStream& s, ResDialog& r )
{
  // 1st we have to look if it's an EXTENDED dialogue resource format
  Q_UINT32 lSignature;
  s >> lSignature;  // really: wDlgVersion (0001) + wSignature (ffff)
  r.dlgheader.bIsExtended = (0xFFFF0000 == (0xFFFF0000 & lSignature));


  if( r.dlgheader.bIsExtended )
  {
    s >> r.dlgheader.lHelpId;
	s >> r.dlgheader.lExtendedStyle;
	s >> r.dlgheader.lStyle;
  }
  else
  {
    r.dlgheader.lStyle = lSignature;
	s >> r.dlgheader.lExtendedStyle;
  }


  // Here we are at normal state again.
  s >> r.dlgheader.numberOfItems;
  s >> r.dlgheader.x;
  s >> r.dlgheader.y;
  s >> r.dlgheader.cx;
  s >> r.dlgheader.cy;
  qDebug( "lStyle = 0x%x, lExtendedStyle = 0x%x, numberOfItems = %d, x = %d, y = %d, cx = %d, cy = %d", r.dlgheader.lStyle, r.dlgheader.lExtendedStyle, r.dlgheader.numberOfItems, r.dlgheader.x, r.dlgheader.y, r.dlgheader.cx, r.dlgheader.cy );


  readNameOrOrdinal( s, r.dlgheader.menuName );
  qDebug( "read menu name %s", r.dlgheader.menuName.toString().data() );


  readNameOrOrdinal( s, r.dlgheader.className );


  qDebug( "read class name %s", r.dlgheader.className.toString().data() );

  readUnicodeString( s, r.dlgheader.caption );


  qDebug( "read caption %s", r.dlgheader.caption.data() );


  // wPointSize and szFontName are only present if FONT statement was used
  if( ( r.dlgheader.lStyle & DS_SETFONT ) == DS_SETFONT ) {
	s >> r.dlgheader.wPointSize;


    if( r.dlgheader.bIsExtended )
    {
      s >> r.dlgheader.wWeight;
      s >> r.dlgheader.bItalic;
      s >> r.dlgheader.bCharset;
    }
	qDebug( "Position after reading point size = %d", s.device()->at() );


//	readUnicodeString( s, r.dlgheader.fontName, true );
	readUnicodeString( s, r.dlgheader.fontName, false );


	qDebug( "read point size %d", r.dlgheader.wPointSize );
	qDebug( "read font name %s", r.dlgheader.fontName.data() );
  }
  else
	r.dlgheader.wPointSize = 0;






  // read all the items
  for( int curitem = 0; curitem < r.dlgheader.numberOfItems; curitem++ ) {

    // align to DWORD boundary for each control
    SkipThePadding( s );


    qDebug( "Starting to read ControlData at %d", s.device()->at() );
	ControlData* cd = new ControlData;


    if( r.dlgheader.bIsExtended )
    {
      s >> r.dlgheader.lHelpId;
      s >> cd->lExtendedStyle;
      s >> cd->lStyle;
    }
    else
    {
	  s >> cd->lStyle;
	  s >> cd->lExtendedStyle;
    }
	s >> cd->x;
	s >> cd->y;
	s >> cd->cx;
	s >> cd->cy;
	s >> cd->wId;


    if( r.dlgheader.bIsExtended )
      SkipThePadding( s );


	readNameOrOrdinal( s, cd->classId );
	readNameOrOrdinal( s, cd->text );
	s >> cd->nExtraStuff;

	r._controls.append( cd );
    QString st = "Read ";
    if( r.dlgheader.bIsExtended )
      st += "extended ";
    st += "dialog item: style = 0x%x, exstyle = 0x%x, x = %d, y = %d, cx = %d,"
          " cy = %d, classId = %s, text = %s, extraData = %d";
    qDebug( st, cd->lStyle, cd->lExtendedStyle, cd->x, cd->y, cd->cx,
			cd->cy, cd->classId.toString().data(), cd->text.toString().data(),
            cd->nExtraStuff );
  }

  return s;
}
