/* --------------------------------------------------------------------------
 * Copyright (C) 2000-2001 Klarlvdalens Datakonsult AB. All rights reserved.
 *
 * This file is part of KNUT, the Windows resource converter for Qt GUI
 * Designer. You may only use and modify it according to the license terms
 * appearing in the file LICENSE included in this package.
 *
 * Holders of a commercial license have additional rights as specified in the
 * license terms shipped with the commercial package.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * ------------------------------------------------------------------------ */


/**
 * Driver for all functions in resconv
 */

#include <stdlib.h>

#include <resconv.h>
#include <qprogressdialog.h>

#include <qapplication.h>
#include <qlistview.h>
#include <qcstring.h>

#include <qnamespace.h>


#ifdef Q_WS_WIN_
#define qstricmp( a,b ) _stricmp( a,b )
#define qstrnicmp( a,b,n ) _strnicmp( a,b,n )
#endif // _QT_WIN32_

const char* cProgressCaption  = "Knut";
const char* cProgressProcInit = "initializing...";

enum FromFormat { FromNone, FromExe, FromRes };
enum ToFormat { ToNone, ToRes, ToSrc, ToXML, ToGUI };


void usage( const char* progname )
{
  qDebug( "Usage:\n" );
  qDebug( "Converting an exe/dll file to a res file:" );
  qDebug( "\t%s --fromexe --tores exefile resfile\n", progname );
  qDebug( "Converting an exe/dll file to an xml file:" );
  qDebug( "\t%s --fromexe --toxml exefile xmlfile\n", progname );
  qDebug( "Converting an exe/dll file to a cpp file:" );
  qDebug( "\t%s --fromexe --tosrc exefile cppfile\n", progname );
  qDebug( "Extracting from an exe/dll file and displaying in listview:" );
  qDebug( "\t%s --fromexe --togui exefile\n", progname );
  qDebug( "Converting a res file to an xml file:" );
  qDebug( "\t%s --fromres --toxml exefile xmlfile\n", progname );
  qDebug( "Converting a res file to a cpp file:" );
  qDebug( "\t%s --fromres --tosrc exefile cppfile\n", progname );
  qDebug( "Extracting from a res file and displaying in listview:" );
  qDebug( "\t%s --fromres --togui exefile\n", progname );

  exit( -2 );
}



void myMessageOutput( QtMsgType, const char* msg )
{}

int main( int argc, char* argv[] )
{
  qInstallMsgHandler( myMessageOutput );

  FromFormat fromFormat = FromNone;
  ToFormat toFormat = ToNone;
  float scalefactor = 1.67;
  QString infilename;
  QString outfilename;
  QString rcfilename( "~/.knutrc" );

  // eval command-line parameters
  QApplication app( argc, argv );
  for( int i = 1; i < argc; i++ ) {
    if( !qstricmp(argv[i], "--fromexe" ) )
	  fromFormat = FromExe;
    else if( !qstricmp(argv[i], "--fromres" ) )
	  fromFormat = FromRes;
    else if( !qstricmp(argv[i], "--tores" ) )
	  toFormat = ToRes;
    else if( !qstricmp(argv[i], "--tosrc" ) )
	  toFormat = ToSrc;
    else if( !qstricmp(argv[i], "--toxml" ) )
	  toFormat = ToXML;
    else if( !qstricmp(argv[i], "--togui" ) )
	  toFormat = ToGUI;
	else if( !qstrnicmp(argv[i], "--scale", 7 ) )
        {
          QString sFactor( argv[i] );
          sFactor.remove(0, 8);
          bool    bOk;
          float   nFactor = sFactor.toFloat( &bOk );
          if( bOk )
            scalefactor = nFactor;
        }
	else if( !qstrnicmp( argv[i], "--fonts:", 7 ) )
        {
          rcfilename = argv[i];
          rcfilename.remove(0, 8);
        }
	else if( argv[i][0] == '-' && strlen( argv[i] ) != 1 )
	  qDebug( "Unknown option: %s", argv[i] );
	else
	  // must be filename
	  if( infilename.isEmpty() )
		infilename = argv[i]; // first filename is infile
	  else
		outfilename = argv[i]; // second (or more) filename is outfile
  };

  // Plausibility checks

  // Both to and from format must be specified
  if( fromFormat == FromNone || toFormat == ToNone )
	usage( argv[0] );

  // Cannot convert from res to res
  if( fromFormat == FromRes && toFormat == ToRes )
	usage( argv[0] );

  // Infile must always be specified
  if( infilename.isEmpty() )
	usage( argv[0] );

  // Outfile must specified unless output format is GUI
  if( toFormat != ToGUI && outfilename.isEmpty() )
	usage( argv[0] );

  // All parameters are correct, call the appropriate converters.
  bool success = false;
  switch( fromFormat ) {
  case FromExe:
    switch( toFormat ) {
    case ToRes:
      success = ResConv::extractToRes( infilename, outfilename, 0 );
      break;
    case ToSrc:
      success = ResConv::extractToSrc( infilename, outfilename );
      break;
    case ToXML:
      {
        QProgressDialog* pd
          = new QProgressDialog( cProgressProcInit, 0, 100,
                                 0, cProgressCaption, true
                                 // ,
                                 //   Qt::WStyle_NoBorder
                                 // | Qt::WStyle_Dialog
                                 // | Qt::WStyle_Title*/
                                 );
        pd->setCaption( cProgressCaption );
        pd->setAutoClose( false );
        pd->setAutoReset( false );
        pd->setMinimumDuration( 0 );
        pd->setProgress( 1 );
        pd->show();
        app.setMainWidget( pd );
        app.processEvents();
        success = ResConv::extractToXML( infilename,  outfilename,
                                         scalefactor, rcfilename, pd );
        pd->setProgress( 100 );
      }
      break;
    case ToGUI: {
        qDebug( "***EXE TO GUI***" );
        QListView* listview = ResConv::extractToGUI( infilename );
        if( listview ) {
          listview->setGeometry( 100, 100, 500, 300 );
          listview->show();
          app.setMainWidget( listview );
          QObject::connect( &app, SIGNAL( lastWindowClosed() ),
                            &app, SLOT( quit() ) );
          app.exec();
          success = true;
        } else
          success = false;
        break;
      }
    default:
      ;
      // cannot happen (ToNone checked above )
  }
  break;
  case FromRes:
	switch( toFormat ) {
	case ToSrc:
	  success = ResConv::convertToSrc( infilename, outfilename );
	  break;
	case ToXML:
	  success = ResConv::convertToXML( infilename, outfilename,
                                       scalefactor, rcfilename, 0 );
	  break;
	case ToGUI: {
	  qDebug( "***RES TO GUI***" );
	  QListView* listview = ResConv::convertToGUI( infilename );
	  if( listview )
      {
		listview->setGeometry( 100, 100, 500, 300 );
		listview->show();
		app.setMainWidget( listview );
		QObject::connect( &app, SIGNAL( lastWindowClosed() ),
						  &app, SLOT( quit() ) );
		app.exec();
		success = true;
	  }
      else
		success = false;
	  break;
	}
	default:
	  ;
	  // cannot happen (ToNone checked above )
	}
	default:
	  ;
	  // cannot happen (FromNone checked above )
  };

  if( !success ) {
	qDebug( "Error while converting/extracting" );
	return -1;
  } else
	return 0;
}
